# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
from collections import Counter

import pytest
from copr.v3 import Client
from flexmock import flexmock

import packit
from packit.config import CommonPackageConfig, aliases
from packit.config.aliases import (
    get_aliases,
    get_all_koji_targets,
    get_branches,
    get_build_targets,
    get_fast_forward_merge_branches_for,
    get_koji_targets,
    get_versions,
)
from packit.copr_helper import CoprHelper
from packit.exceptions import PackitException
from tests.spellbook import ALL_KOJI_TARGETS_SNAPSHOT


@pytest.mark.usefixtures("mock_get_aliases")
class TestGetVersions:
    @pytest.mark.parametrize(
        "name,versions",
        [
            ("fedora-29", {"fedora-29"}),
            ("epel-8", {"epel-8"}),
            ("fedora-rawhide", {"fedora-rawhide"}),
            ("openmandriva-rolling", {"openmandriva-rolling"}),
            ("opensuse-leap-15.0", {"opensuse-leap-15.0"}),
            ("fedora-stable", {"fedora-31", "fedora-32"}),
            ("fedora-development", {"fedora-33", "fedora-rawhide"}),
            ("fedora-all", {"fedora-31", "fedora-32", "fedora-33", "fedora-rawhide"}),
            ("centos-stream-8", {"centos-stream-8"}),
        ],
    )
    def test_get_versions(self, name, versions, mock_get_aliases):
        assert get_versions(name) == versions

    @pytest.mark.parametrize(
        "names,versions",
        [
            (
                ["fedora-30", "fedora-stable"],
                {"fedora-30", "fedora-31", "fedora-32"},
            ),
            (["fedora-31", "fedora-stable"], {"fedora-31", "fedora-32"}),
            ([], {"fedora-31", "fedora-32"}),
        ],
    )
    def test_get_versions_from_multiple_values(self, names, versions):
        assert get_versions(*names) == versions

    def test_get_versions_empty_without_default(self):
        assert get_versions(default=None) == set()


@pytest.mark.usefixtures("mock_get_aliases")
class TestGetBuildTargets:
    @pytest.mark.parametrize(
        "name,targets",
        [
            ("rawhide", {"fedora-rawhide-x86_64"}),
            ("fedora-29", {"fedora-29-x86_64"}),
            ("epel-8", {"epel-8-x86_64"}),
            ("fedora-rawhide", {"fedora-rawhide-x86_64"}),
            ("openmandriva-rolling", {"openmandriva-rolling-x86_64"}),
            ("opensuse-leap-15.0", {"opensuse-leap-15.0-x86_64"}),
            ("centos-stream", {"centos-stream-8-x86_64"}),
            ("centos-stream-x86_64", {"centos-stream-8-x86_64"}),
            ("centos-stream-8", {"centos-stream-8-x86_64"}),
            ("centos-stream-8-x86_64", {"centos-stream-8-x86_64"}),
            ("fedora-stable", {"fedora-31-x86_64", "fedora-32-x86_64"}),
            ("fedora-development", {"fedora-33-x86_64", "fedora-rawhide-x86_64"}),
            ("fedora-29-x86_64", {"fedora-29-x86_64"}),
            ("fedora-29-aarch64", {"fedora-29-aarch64"}),
            ("fedora-29-i386", {"fedora-29-i386"}),
            (
                "fedora-stable-aarch64",
                {"fedora-31-aarch64", "fedora-32-aarch64"},
            ),
            (
                "fedora-development-aarch64",
                {"fedora-33-aarch64", "fedora-rawhide-aarch64"},
            ),
            (
                "fedora-all",
                {
                    "fedora-31-x86_64",
                    "fedora-32-x86_64",
                    "fedora-33-x86_64",
                    "fedora-rawhide-x86_64",
                },
            ),
        ],
    )
    def test_get_build_targets(self, name, targets, mock_get_aliases):
        assert get_build_targets(name) == targets

    def test_get_build_targets_invalid_input(self):
        name = "rafhajd"
        with pytest.raises(PackitException) as ex:
            get_build_targets(name)
        err_msg = (
            f"Cannot get build target from '{name}'"
            f", packit understands values like these: "
        )
        assert err_msg in str(ex.value)

    def test_get_build_targets_without_default(self):
        assert get_build_targets(default="") == set()

    @pytest.mark.parametrize(
        "names,versions",
        [
            (
                ["fedora-30", "fedora-stable"],
                {"fedora-30-x86_64", "fedora-31-x86_64", "fedora-32-x86_64"},
            ),
            (
                ["fedora-31", "fedora-stable"],
                {"fedora-31-x86_64", "fedora-32-x86_64"},
            ),
        ],
    )
    def test_get_build_targets_from_multiple_values(self, names, versions):
        assert get_build_targets(*names) == versions


class TestGetBranches:
    @pytest.mark.parametrize(
        "name,default_dg_branch,branches",
        [
            ("fedora-29", None, {"f29"}),
            ("fedora-rawhide", None, {"main"}),
            ("fedora-rawhide", "main", {"main"}),
            ("fedora-rawhide", "master", {"master"}),
            ("rawhide", None, {"main"}),
            ("rawhide", "main", {"main"}),
            ("rawhide", "master", {"master"}),
            ("main", None, {"main"}),
            ("master", None, {"master"}),
            ("f30", None, {"f30"}),
            ("fedora-development", None, {"f33", "main"}),
            ("fedora-development", "main", {"f33", "main"}),
            ("fedora-development", "master", {"f33", "master"}),
            ("fedora-stable", None, {"f31", "f32"}),
            ("epel-7", None, {"epel7"}),
            ("epel7", None, {"epel7"}),
            ("el6", None, {"el6"}),
            ("epel-6", None, {"el6"}),
            ("fedora-all", None, {"f31", "f32", "f33", "main"}),
            ("fedora-all", "main", {"f31", "f32", "f33", "main"}),
            ("fedora-all", "master", {"f31", "f32", "f33", "master"}),
        ],
    )
    def test_get_branches(self, name, default_dg_branch, branches, mock_get_aliases):
        if default_dg_branch:
            assert get_branches(name, default_dg_branch=default_dg_branch) == branches
        else:
            assert get_branches(name) == branches

    @pytest.mark.parametrize(
        "names,versions",
        [
            (["fedora-30", "fedora-stable"], {"f30", "f31", "f32"}),
            (["fedora-31", "fedora-stable"], {"f31", "f32"}),
        ],
    )
    def test_get_branches_from_multiple_values(self, names, versions):
        flexmock(packit.config.aliases).should_receive("get_versions").and_return(
            versions,
        )
        assert get_branches(*names) == versions

    @pytest.mark.parametrize(
        "names,versions",
        [
            (
                {
                    "fedora-30": {"fast_forward_merge_into": ["f31"]},
                    "fedora-stable": {},
                },
                {"f30", "f32"},
            ),
        ],
    )
    def test_get_branches_from_multiple_values_in_dict(self, names, versions):
        flexmock(packit.config.aliases).should_receive("get_versions").and_return(
            versions,
        )
        assert get_branches(*names) == versions

    def test_get_branches_without_default(self):
        assert get_branches(default=None) == set()

    @pytest.mark.parametrize(
        "config,branches,ff_branches",
        [
            (
                CommonPackageConfig(
                    dist_git_branches={
                        "rawhide": {"fast_forward_merge_into": ["f33"]},
                        "f35": {},
                        "f34": {},
                    },
                ),
                {"main", "f35", "f34"},
                {"rawhide": {"f33"}, "main": {"f33"}, "f35": set(), "f34": set()},
            ),
            (
                CommonPackageConfig(
                    # no sense but possible!
                    dist_git_branches={
                        "fedora-branched": {
                            "fast_forward_merge_into": ["fedora-stable"],
                        },
                    },
                ),
                {"f39", "f40"},
                {"f39": {"f39", "f40"}, "f40": {"f39", "f40"}},
            ),
        ],
    )
    def test_get_fast_forward_merge_branches_for(
        self,
        config,
        branches,
        ff_branches,
    ):
        mock_aliases_module = flexmock(packit.config.aliases)
        mock_aliases_module.should_receive("get_aliases").and_return(
            {
                "fedora-all": ["fedora-39", "fedora-40", "fedora-rawhide"],
                "fedora-stable": ["fedora-39", "fedora-40"],
                "fedora-development": ["fedora-rawhide"],
                "fedora-latest": ["fedora-40"],
                "fedora-latest-stable": ["fedora-40"],
                "fedora-branched": ["fedora-39", "fedora-40"],
                "epel-all": ["epel-8", "epel-9"],
            },
        )

        assert branches == get_branches(*config.dist_git_branches)
        for source_branch in get_branches(*config.dist_git_branches, with_aliases=True):
            assert (
                get_fast_forward_merge_branches_for(
                    config.dist_git_branches,
                    source_branch,
                )
                == ff_branches[source_branch]
            )


class TestGetKojiTargets:
    @pytest.mark.parametrize("target", ALL_KOJI_TARGETS_SNAPSHOT)
    def test_preserve_koji_targets_single(self, target):
        assert {target} == get_koji_targets(target)

    def test_preserve_all_koji_targets_together(self):
        assert set(ALL_KOJI_TARGETS_SNAPSHOT) == get_koji_targets(
            *ALL_KOJI_TARGETS_SNAPSHOT,
        )

    @pytest.mark.parametrize(
        "name,targets",
        [
            ("fedora-29", {"f29"}),
            ("fedora-rawhide", {"rawhide"}),
            ("rawhide", {"rawhide"}),
            ("master", {"master"}),
            ("f30", {"f30"}),
            ("fedora-development", {"f33", "rawhide"}),
            ("fedora-stable", {"f31", "f32"}),
            ("epel-7", {"epel7"}),
            ("epel7", {"epel7"}),
            ("el6", {"epel6"}),
            ("epel-6", {"epel6"}),
            ("fedora-all", {"f31", "f32", "f33", "rawhide"}),
            ("epel-all", {"epel6", "epel7", "epel8"}),
        ],
    )
    def test_get_koji_targets(self, name, targets, mock_get_aliases):
        assert get_koji_targets(name) == targets

    def test_get_koji_targets_without_default(self):
        assert get_koji_targets(default=None) == set()


class TestGetAllKojiTargets:
    def test_get_all_koji_targets(self):
        targets = get_all_koji_targets()
        assert targets
        assert "Name" not in targets
        assert "rawhide" in targets
        assert "epel8" in targets
        assert all(isinstance(target, str) for target in targets)


class TestGetAliases:
    @pytest.mark.parametrize(
        "releases_list, expected_return",
        [
            pytest.param(
                [
                    ("F30", "Fedora 30", "FEDORA", "archived"),
                    ("F31", "Fedora 31", "FEDORA", "archived"),
                    ("F32", "Fedora 32", "FEDORA", "current"),
                    ("F33", "Fedora 33", "FEDORA", "current"),
                    ("F34", "Fedora 34", "FEDORA", "frozen"),
                    ("F35", "Fedora 35", "FEDORA", "pending"),
                    ("F31F", "Fedora 31 Flatpaks", "FEDORA-FLATPAK", "current"),
                    ("EPEL-8", "Fedora EPEL 8", "FEDORA-EPEL", "current"),
                ],
                {
                    "fedora-all": [
                        "fedora-32",
                        "fedora-33",
                        "fedora-34",
                        "fedora-rawhide",
                    ],
                    "fedora-stable": ["fedora-32", "fedora-33"],
                    "fedora-development": ["fedora-34", "fedora-rawhide"],
                    "fedora-latest": ["fedora-34"],
                    "fedora-latest-stable": ["fedora-33"],
                    "fedora-branched": ["fedora-32", "fedora-33", "fedora-34"],
                    "epel-all": ["epel-8"],
                },
                id="after_branching",
            ),
            pytest.param(
                [
                    ("F30", "Fedora 30", "FEDORA", "archived"),
                    ("F31", "Fedora 31", "FEDORA", "archived"),
                    ("F32", "Fedora 32", "FEDORA", "current"),
                    ("F33", "Fedora 33", "FEDORA", "current"),
                    ("F34", "Fedora 34", "FEDORA", "current"),
                    ("F35", "Fedora 35", "FEDORA", "pending"),
                    ("F31F", "Fedora 31 Flatpaks", "FEDORA-FLATPAK", "current"),
                    ("EPEL-8", "Fedora EPEL 8", "FEDORA-EPEL", "current"),
                ],
                {
                    "fedora-all": [
                        "fedora-32",
                        "fedora-33",
                        "fedora-34",
                        "fedora-rawhide",
                    ],
                    "fedora-stable": ["fedora-32", "fedora-33", "fedora-34"],
                    "fedora-development": ["fedora-rawhide"],
                    "fedora-latest": ["fedora-34"],
                    "fedora-latest-stable": ["fedora-34"],
                    "fedora-branched": ["fedora-32", "fedora-33", "fedora-34"],
                    "epel-all": ["epel-8"],
                },
                id="after_release",
            ),
            pytest.param(
                [
                    ("F30", "Fedora 30", "FEDORA", "archived"),
                    ("F31", "Fedora 31", "FEDORA", "archived"),
                    ("F32", "Fedora 32", "FEDORA", "archived"),
                    ("F33", "Fedora 33", "FEDORA", "current"),
                    ("F34", "Fedora 34", "FEDORA", "current"),
                    ("F35", "Fedora 35", "FEDORA", "pending"),
                    ("F31F", "Fedora 31 Flatpaks", "FEDORA-FLATPAK", "current"),
                    ("EPEL-8", "Fedora EPEL 8", "FEDORA-EPEL", "current"),
                ],
                {
                    "fedora-all": [
                        "fedora-33",
                        "fedora-34",
                        "fedora-rawhide",
                    ],
                    "fedora-stable": ["fedora-33", "fedora-34"],
                    "fedora-development": ["fedora-rawhide"],
                    "fedora-latest": ["fedora-34"],
                    "fedora-latest-stable": ["fedora-34"],
                    "fedora-branched": ["fedora-33", "fedora-34"],
                    "epel-all": ["epel-8"],
                },
                id="after_eol",
            ),
        ],
    )
    def test_get_aliases(self, releases_list, expected_return, bodhi_client_response):
        bodhi_instance_mock = flexmock()
        bodhi_instance_mock.should_receive("get_releases").and_return(
            bodhi_client_response(releases_list),
        )
        flexmock(aliases).should_receive("get_bodhi_client").and_return(
            bodhi_instance_mock,
        ).once()

        get_aliases.cache_clear()
        aliases_result = get_aliases()

        assert Counter(aliases_result["fedora-stable"]) == Counter(
            expected_return["fedora-stable"],
        )
        assert Counter(aliases_result["fedora-development"]) == Counter(
            expected_return["fedora-development"],
        )
        assert Counter(aliases_result["fedora-all"]) == Counter(
            expected_return["fedora-all"],
        )
        assert Counter(aliases_result["fedora-latest"]) == Counter(
            expected_return["fedora-latest"],
        )
        assert Counter(aliases_result["fedora-latest-stable"]) == Counter(
            expected_return["fedora-latest-stable"],
        )
        assert Counter(aliases_result["fedora-branched"]) == Counter(
            expected_return["fedora-branched"],
        )
        assert Counter(aliases_result["epel-all"]) == Counter(
            expected_return["epel-all"],
        )


@pytest.mark.parametrize(
    "targets, chroots, expected_result",
    [
        pytest.param(["f1", "f2"], ["f1", "f2"], {"f1", "f2"}, id="identical"),
        pytest.param(["f1", "f2"], ["f2", "f3"], {"f2"}, id="some_common"),
        pytest.param(["f1", "f2"], ["f3", "f4"], set(), id="none_common"),
        pytest.param([], ["f1", "f2"], set(), id="one_empty"),
        pytest.param([], [], set(), id="both_empty"),
    ],
)
def test_get_valid_build_targets(targets, chroots, expected_result):
    copr_helper = CoprHelper(flexmock())
    flexmock(Client).should_receive("create_from_config_file")
    flexmock(packit.config.aliases).should_receive("get_build_targets").and_return(
        targets,
    )
    flexmock(CoprHelper).should_receive("get_available_chroots").and_return(chroots)

    assert copr_helper.get_valid_build_targets(*targets) == expected_result


@pytest.mark.parametrize(
    "name, default",
    [
        pytest.param(["f1", "f2"], {"default": "test"}, id="name_set-default_set"),
        pytest.param(["f1", "f2"], {"default": "None"}, id="name_set-default_None"),
        pytest.param([], {"default": "test"}, id="name_None-default_set"),
        pytest.param([], {"default": None}, id="name_None-default_None"),
    ],
)
def test_get_valid_build_targets_get_aliases_call(name, default):
    flexmock(packit.config.aliases).should_receive("get_build_targets").with_args(
        *name,
        **default,
    ).and_return(set())
    flexmock(CoprHelper).should_receive("get_available_chroots").and_return(set())
    CoprHelper(flexmock()).get_valid_build_targets(*name, **default)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import NamedTuple


class SourcesItem(NamedTuple):
    path: str
    url: str

    def __repr__(self):
        return f"SourcesItem(path={self.path}, url={self.url})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SourcesItem):
            raise NotImplementedError()

        return self.path == other.path and self.url == other.url



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""Check synchronization status of source-git and dist-git"""

import pathlib

import click

from packit.api import PackitAPI
from packit.cli.utils import cover_packit_exception, iterate_packages_source_git
from packit.config import Config, PackageConfig, pass_config
from packit.local_project import CALCULATE, LocalProjectBuilder


@click.command("status")
@click.argument("source-git", type=click.Path(exists=True, file_okay=False))
@click.argument("dist-git", type=click.Path(exists=True, file_okay=False))
@pass_config
@cover_packit_exception
@iterate_packages_source_git
def source_git_status(
    config: Config,
    package_config: PackageConfig,
    source_git: str,
    dist_git: str,
):
    """Tell the synchronization status of a source-git and a dist-git repo.

    This command checks the commit history in the provided source-git
    and dist-git repos and informs about the range of commits to be
    synchronized from dist-git to source-git or the other way around,
    or informs that the repositories are in sync.

    If possible, the status command also provides instructions on how
    to synchronize the repositories.
    """
    source_git_path = pathlib.Path(source_git).resolve()
    dist_git_path = pathlib.Path(dist_git).resolve()
    builder = LocalProjectBuilder(offline=True)
    api = PackitAPI(
        config=config,
        package_config=package_config,
        upstream_local_project=builder.build(
            working_dir=source_git_path,
            git_repo=CALCULATE,
        ),
        downstream_local_project=builder.build(
            working_dir=dist_git_path,
            git_repo=CALCULATE,
        ),
    )
    click.echo(api.sync_status_string(source_git=source_git, dist_git=dist_git))



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.exceptions import PackitLookasideCacheException
from packit.utils.lookaside import LookasideCache, configparser, pyrpkg


@pytest.mark.parametrize(
    "config, sources, package, result",
    [
        (
            {
                "lookaside": "https://src.fedoraproject.org/repo/pkgs",
                "lookaside_cgi": "https://src.fedoraproject.org/repo/pkgs/upload.cgi",
                "lookasidehash": "sha512",
            },
            [
                {
                    "file": "packitos-0.57.0.tar.gz",
                    "hash": "27e4f97e262d7b1eb0af79ef9ea8ceae"
                    "d024dfe7b3d7ca0141d63195e7a9d6ee"
                    "a147d8eef919cd7919435abc5b729ca0"
                    "4d9800a9df1c0334c6ca42a5747a8755",
                    "hashtype": "sha512",
                },
            ],
            "packit",
            [
                {
                    "path": "packitos-0.57.0.tar.gz",
                    "url": "https://src.fedoraproject.org/repo/pkgs"
                    "/packit/packitos-0.57.0.tar.gz/sha512/"
                    "27e4f97e262d7b1eb0af79ef9ea8ceae"
                    "d024dfe7b3d7ca0141d63195e7a9d6ee"
                    "a147d8eef919cd7919435abc5b729ca0"
                    "4d9800a9df1c0334c6ca42a5747a8755"
                    "/packitos-0.57.0.tar.gz",
                },
            ],
        ),
        (
            {
                "lookaside": "https://sources.stream.centos.org/sources",
                "lookaside_cgi": "https://sources.stream.rdu2.redhat.com/lookaside/upload.cgi",
                "lookasidehash": "sha512",
                "lookaside_namespaced": True,
            },
            [
                {
                    "file": "man-pages-5.10.tar.xz",
                    "hash": "a23f90136b0bf471f5ae3917ae0e558f"
                    "ec0671cace8ccdd8e244f41f11fefa4a"
                    "c0df84cf972cc20a1792d7b930db5e2c"
                    "451881c0937edabf7d5e1ec46c4760ed",
                    "hashtype": "sha512",
                },
                {
                    "file": "man-pages-additional-20140218.tar.xz",
                    "hash": "c7874db32a9bdefaea6c6be6549e6e65"
                    "38fa1d93260bf342dd0d9821fa05754a"
                    "a79a723e701493c81b2e1f460918429e"
                    "b9b5edb704b55878b1e5ed585a3ff07d",
                    "hashtype": "sha512",
                },
                {
                    "file": "man-pages-posix-2017-a.tar.xz",
                    "hash": "dac6bd5bb3e1d5f8918bad3eb15e08ee"
                    "b3e06ae160c04ccd5619bfb0c536139a"
                    "c06faa62b6856656a1bb9a7496f3148e"
                    "52a5227b83e4099be6e6b93230de211d",
                    "hashtype": "sha512",
                },
            ],
            "man-pages",
            [
                {
                    "path": "man-pages-5.10.tar.xz",
                    "url": "https://sources.stream.centos.org/sources"
                    "/rpms/man-pages/man-pages-5.10.tar.xz/sha512/"
                    "a23f90136b0bf471f5ae3917ae0e558f"
                    "ec0671cace8ccdd8e244f41f11fefa4a"
                    "c0df84cf972cc20a1792d7b930db5e2c"
                    "451881c0937edabf7d5e1ec46c4760ed"
                    "/man-pages-5.10.tar.xz",
                },
                {
                    "path": "man-pages-additional-20140218.tar.xz",
                    "url": "https://sources.stream.centos.org/sources"
                    "/rpms/man-pages/man-pages-additional-20140218.tar.xz/sha512/"
                    "c7874db32a9bdefaea6c6be6549e6e65"
                    "38fa1d93260bf342dd0d9821fa05754a"
                    "a79a723e701493c81b2e1f460918429e"
                    "b9b5edb704b55878b1e5ed585a3ff07d"
                    "/man-pages-additional-20140218.tar.xz",
                },
                {
                    "path": "man-pages-posix-2017-a.tar.xz",
                    "url": "https://sources.stream.centos.org/sources"
                    "/rpms/man-pages/man-pages-posix-2017-a.tar.xz/sha512/"
                    "dac6bd5bb3e1d5f8918bad3eb15e08ee"
                    "b3e06ae160c04ccd5619bfb0c536139a"
                    "c06faa62b6856656a1bb9a7496f3148e"
                    "52a5227b83e4099be6e6b93230de211d"
                    "/man-pages-posix-2017-a.tar.xz",
                },
            ],
        ),
        ({}, [], "test", []),
    ],
)
def test_get_lookaside_sources(config, sources, package, result):
    flexmock(
        configparser,
        ConfigParser=lambda: flexmock(
            read=lambda _: None,
            items=lambda _, **__: config,
        ),
    )
    flexmock(
        pyrpkg.sources,
        SourcesFile=lambda *_: flexmock(entries=[flexmock(**s) for s in sources]),
    )
    if "lookaside" not in config:
        with pytest.raises(PackitLookasideCacheException):
            LookasideCache("").get_sources("", package)
    else:
        assert LookasideCache("").get_sources("", package) == result



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from packit.utils.commands import cwd
from tests.spellbook import call_packit


def test_init_pass(upstream_without_config):
    with cwd(upstream_without_config):
        assert not (upstream_without_config / ".packit.yaml").is_file()

        # This test requires packit on pythonpath
        result = call_packit(parameters=["init"])

        assert result.exit_code == 0

        assert (upstream_without_config / ".packit.yaml").is_file()


def test_init_fail(cwd_upstream_or_distgit):
    result = call_packit(parameters=["init"], working_dir=str(cwd_upstream_or_distgit))

    assert result.exit_code == 2  # packit config already exists --force needed


def test_init_search_for_specfile_root(upstream_without_config):
    with cwd(upstream_without_config):
        (upstream_without_config / "file.spec").touch()
        call_packit(parameters=["init"])

        with open(upstream_without_config / ".packit.yaml") as f:
            assert "\nspecfile_path: file.spec\n" in f.read()


def test_init_search_for_specfile_recursive(upstream_without_config):
    with cwd(upstream_without_config):
        nested_path = upstream_without_config / "awesome_directory" / "nested"
        nested_path.mkdir(parents=True, exist_ok=True)
        (nested_path / "pickle.spec").touch()
        call_packit(parameters=["init"])

        with open(upstream_without_config / ".packit.yaml") as f:
            assert "\nspecfile_path: awesome_directory/nested/pickle.spec\n" in f.read()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import time
from datetime import datetime, timedelta
from typing import Any, Callable, Optional

from cachetools.func import ttl_cache
from copr.v3 import Client as CoprClient
from copr.v3.exceptions import (
    CoprAuthException,
    CoprException,
    CoprNoResultException,
    CoprRequestException,
)
from munch import Munch

from packit.config import aliases  # so we can mock in tests
from packit.config.aliases import get_build_targets
from packit.config.common_package_config import MockBootstrapSetup
from packit.constants import CHROOT_SPECIFIC_COPR_CONFIGURATION, COPR2GITHUB_STATE
from packit.exceptions import PackitCoprProjectException, PackitCoprSettingsException
from packit.local_project import LocalProject

logger = logging.getLogger(__name__)


class CoprHelper:
    def __init__(self, upstream_local_project: LocalProject) -> None:
        self.upstream_local_project = upstream_local_project
        self._copr_client = None

    def __repr__(self):
        return (
            "CoprHelper("
            f"upstream_local_project='{self.upstream_local_project}', "
            f"copr_client='{self.copr_client}')"
        )

    def get_copr_client(self) -> CoprClient:
        """Not static because of the flex-mocking."""
        return CoprClient.create_from_config_file()

    @property
    def copr_client(self) -> CoprClient:
        if self._copr_client is None:
            self._copr_client = self.get_copr_client()
        return self._copr_client

    @property
    def configured_owner(self) -> Optional[str]:
        return self.copr_client.config.get("username")

    def copr_web_build_url(self, build: Munch) -> str:
        """Construct web frontend url because build.repo_url is not much user-friendly."""
        copr_url = self.copr_client.config.get("copr_url")
        return f"{copr_url}/coprs/build/{build.id}/"

    def get_copr_settings_url(
        self,
        owner: str,
        project: str,
        section: Optional[str] = None,
    ):
        copr_url = self.copr_client.config.get("copr_url")
        section = section or "edit"

        # COPR groups starts with '@' but url have '/g/owner'
        if owner.startswith("@"):
            owner = f"g/{owner[1:]}"

        return f"{copr_url}/coprs/{owner}/{project}/{section}/"

    def get_valid_build_targets(
        self,
        *name: str,
        default: Optional[str] = aliases.DEFAULT_VERSION,
    ) -> set:
        """
        For the provided iterable of names, expand them using get_build_targets() into valid
        Copr chhroot names and intersect this set with current available Copr chroots.

        :param name: name(s) of the system and version or target name. (passed to
                    packit.config.aliases.get_build_targets() function)
                or target name (e.g. "fedora-30-x86_64" or "fedora-stable-x86_64")
        :param default: used if no positional argument was given
        :return: set of build targets available also in copr chroots
        """
        build_targets = aliases.get_build_targets(*name, default=default)
        logger.info(f"Build targets: {build_targets} ")
        copr_chroots = self.get_available_chroots()
        logger.info(f"Copr chroots: {copr_chroots} ")
        logger.info(f"Result set: {set(build_targets) & set(copr_chroots)}")
        return set(build_targets) & set(copr_chroots)

    def _get_chroot_specific_configuration_to_update(
        self,
        project: str,
        owner: Optional[str] = None,
        targets_dict: Optional[dict] = None,  # chroot specific configuration
    ) -> dict[str, dict[str, tuple]]:
        """
        Using the provided targets_dict, update chroot specific configuration
        """
        if not targets_dict:
            return {}

        update_dict = {}
        # let's get the chroot specific configuration to update
        for target, chroot_configuration in targets_dict.items():
            chroot_names = get_build_targets(target)
            for chroot_name in chroot_names:
                if set(chroot_configuration.keys()).intersection(
                    CHROOT_SPECIFIC_COPR_CONFIGURATION.keys(),
                ):
                    logger.info(
                        f"There is chroot-specific configuration for {chroot_name}",
                    )
                    # only update when needed
                    try:
                        copr_chroot_configuration = (
                            self.copr_client.project_chroot_proxy.get(
                                ownername=owner,
                                projectname=project,
                                chrootname=chroot_name,
                            )
                        )
                    except CoprNoResultException:
                        logger.debug(
                            "It was not possible to get chroot configuration for "
                            f"{chroot_name}",
                        )
                        continue

                    update_dict_chroot = {}
                    for c, default in CHROOT_SPECIFIC_COPR_CONFIGURATION.items():
                        if (
                            old_value := copr_chroot_configuration.get(
                                c,
                                default,
                            )
                        ) != (new_value := chroot_configuration.get(c, default)):
                            update_dict_chroot[c] = (old_value, new_value)
                    if update_dict_chroot:
                        update_dict[chroot_name] = update_dict_chroot
        return update_dict

    def _update_chroot_specific_configuration(
        self,
        owner: str,
        project: str,
        update_dict: dict[str, dict[str, tuple]],
    ):
        for chroot, update_dict_chroot in update_dict.items():
            diff_string = [
                f"{field}: {old} -> {new}"
                for field, (old, new) in update_dict_chroot.items()
            ]
            logger.info(
                f"Update {owner}/{project} {chroot}: {diff_string}",
            )
            update_args = {
                field: new_value
                for field, (old_value, new_value) in update_dict_chroot.items()
            }
            self.copr_client.project_chroot_proxy.edit(
                ownername=owner,
                projectname=project,
                chrootname=chroot,
                **update_args,
            )

    def create_or_update_copr_project(
        self,
        project: str,
        chroots: list[str],
        owner: Optional[str] = None,
        description: Optional[str] = None,
        instructions: Optional[str] = None,
        list_on_homepage: Optional[bool] = False,
        preserve_project: Optional[bool] = False,
        additional_packages: Optional[list[str]] = None,
        additional_repos: Optional[list[str]] = None,
        bootstrap: Optional[MockBootstrapSetup] = None,
        request_admin_if_needed: bool = False,
        targets_dict: Optional[dict] = None,  # chroot specific configuration
        module_hotfixes: bool = False,
        follow_fedora_branching: bool = False,
    ) -> None:
        """
        Create or update a project in copr.

        Raises PackitCoprException on any problems.
        """
        default_description = (
            "Continuous builds initiated by Packit service.\n"
            "For more info check out https://packit.dev/"
        )

        default_instructions = (
            (
                "You can check out the upstream project "
                f"{self.upstream_local_project.git_url} to find out how to consume these builds. "
                f"This copr project is created and handled by the Packit project "
                "(https://packit.dev/)."
            )
            if self.upstream_local_project
            else None
        )

        delete_after_days: Optional[int] = (
            None if preserve_project is None else -1 if preserve_project else 60
        )

        logger.info(f"Creating {owner}/{project} Copr project.")
        try:
            copr_proj = self.copr_client.project_proxy.add(
                ownername=owner,
                projectname=project,
                chroots=chroots,
                description=description or default_description,
                contact="https://github.com/packit/packit/issues",
                # don't show project on Copr homepage by default
                unlisted_on_hp=not list_on_homepage,
                # delete project after the specified period of time
                delete_after_days=delete_after_days,
                additional_repos=additional_repos,
                bootstrap=bootstrap.value if bootstrap is not None else None,
                instructions=instructions or default_instructions,
                module_hotfixes=module_hotfixes,
                follow_fedora_branching=follow_fedora_branching,
                exist_ok=True,
            )
        except (CoprException, CoprRequestException) as ex:
            response = ex.result.__response__
            if response.status_code >= 500:
                error = (
                    f"Packit received HTTP {response.status_code} {response.reason} "
                    "from Copr Service. "
                    "Check the Copr status page: https://copr.fedorainfracloud.org/status/stats/, "
                    "or ask for help in Fedora Build System matrix channel: "
                    "https://matrix.to/#/#buildsys:fedoraproject.org."
                )
                logger.debug(
                    f"Unexpected Copr error: {response.status_code} {response.reason}: "
                    f"{response.text}",
                )
            else:
                error = (
                    f"Cannot create a new Copr project "
                    f"(owner={owner} project={project} chroots={chroots}): {ex}. "
                    f"Copr HTTP response is {response.status_code} {response.reason}."
                )
            logger.error(error)
            logger.error(ex.result)
            raise PackitCoprProjectException(error) from ex

        fields_to_change = self.get_fields_to_change(
            copr_proj=copr_proj,
            additional_repos=additional_repos,
            chroots=chroots,
            description=description,
            instructions=instructions,
            list_on_homepage=list_on_homepage,
            delete_after_days=delete_after_days,
            module_hotfixes=module_hotfixes,
            bootstrap=bootstrap,
        )
        try:
            if fields_to_change:
                failure_message = (
                    f"Copr project update failed for '{owner}/{project}' project."
                )
                self.update_copr_project(owner, project, fields_to_change)

            failure_message = (
                f"Copr project chroot configuration update failed "
                f"for '{owner}/{project}' project."
            )
            chroot_specific_config_to_update = (
                self._get_chroot_specific_configuration_to_update(
                    project,
                    owner,
                    targets_dict,
                )
            )

            # transform the dict for the user message purposes
            fields_to_change = {
                f"{chroot}: {field}": values
                for chroot, chroot_config in chroot_specific_config_to_update.items()
                for field, values in chroot_config.items()
            }

            self._update_chroot_specific_configuration(
                project=project,
                owner=owner,
                update_dict=chroot_specific_config_to_update,
            )

        except CoprAuthException as ex:
            if "Only owners and admins may update their projects." in str(ex):
                if request_admin_if_needed:
                    logger.info(
                        "Admin permissions are required "
                        "in order to be able to edit project settings. "
                        "Requesting the admin rights for the "
                        f"copr '{owner}/{project}' project.",
                    )
                    self.copr_client.project_proxy.request_permissions(
                        ownername=owner,
                        projectname=project,
                        permissions={"admin": True},
                    )
                else:
                    logger.warning(
                        f"Admin permissions are required for copr '{owner}/{project}' project"
                        f"in order to be able to edit project settings. "
                        f"You can make a request by specifying --request-admin-if-needed "
                        f"when using Packit CLI.",
                    )
            raise PackitCoprSettingsException(
                failure_message,
                fields_to_change=fields_to_change,
            ) from ex

    def update_copr_project(
        self,
        owner: str,
        project: str,
        fields_to_change: dict[str, tuple],
    ):
        logger.info(f"Updating copr project '{owner}/{project}'")
        for field, (old, new) in fields_to_change.items():
            logger.debug(f"{field}: {old} -> {new}")
            kwargs: dict[str, Any] = {
                arg_name: new for arg_name, (old, new) in fields_to_change.items()
            }
            logger.debug(f"Copr edit arguments: {kwargs}")
            self.copr_client.project_proxy.edit(
                ownername=owner,
                projectname=project,
                **kwargs,
            )

    def get_fields_to_change(
        self,
        copr_proj,
        additional_repos: Optional[list[str]] = None,
        chroots: Optional[list[str]] = None,
        description: Optional[str] = None,
        instructions: Optional[str] = None,
        list_on_homepage: Optional[bool] = True,
        delete_after_days: Optional[int] = None,
        module_hotfixes: Optional[bool] = False,
        bootstrap: Optional[MockBootstrapSetup] = None,
    ) -> dict[str, tuple[Any, Any]]:
        fields_to_change: dict[str, tuple[Any, Any]] = {}
        if chroots is not None:
            old_chroots = self.get_chroots(copr_project=copr_proj)

            new_chroots = None
            if not set(chroots).issubset(old_chroots):
                new_chroots = list(set(chroots) | old_chroots)

            if new_chroots:
                new_chroots.sort()
                fields_to_change["chroots"] = (
                    list(old_chroots),
                    new_chroots,
                )
        if description and copr_proj.description != description:
            fields_to_change["description"] = (copr_proj.description, description)

        if instructions:
            if "instructions" not in copr_proj:
                logger.debug(
                    "The `instructions` key was not received from Copr. "
                    "We can't check that value to see if the update is needed.",
                )
            elif copr_proj.instructions != instructions:
                fields_to_change["instructions"] = (
                    copr_proj.instructions,
                    instructions,
                )

        if list_on_homepage is not None:
            if "unlisted_on_hp" not in copr_proj:
                logger.debug(
                    "The `unlisted_on_hp` key was not received from Copr. "
                    "We can't check that value to see if the update is needed.",
                )
            elif copr_proj.unlisted_on_hp != (not list_on_homepage):
                fields_to_change["unlisted_on_hp"] = (
                    copr_proj.unlisted_on_hp,
                    (not list_on_homepage),
                )

        if delete_after_days is not None:
            if "delete_after_days" not in copr_proj:
                logger.debug(
                    "The `delete_after_days` key was not received from Copr. "
                    "We can't check that value to see if the update is needed.",
                )
            elif copr_proj.delete_after_days != delete_after_days:
                fields_to_change["delete_after_days"] = (
                    copr_proj.delete_after_days,
                    delete_after_days,
                )

        if additional_repos is not None and set(copr_proj.additional_repos) != set(
            additional_repos,
        ):
            fields_to_change["additional_repos"] = (
                copr_proj.additional_repos,
                additional_repos,
            )

        if module_hotfixes is not None and copr_proj.module_hotfixes != module_hotfixes:
            fields_to_change["module_hotfixes"] = (
                copr_proj.module_hotfixes,
                module_hotfixes,
            )

        if bootstrap is not None and copr_proj.bootstrap != bootstrap.value:
            fields_to_change["bootstrap"] = (
                copr_proj.bootstrap,
                bootstrap.value,
            )

        return fields_to_change

    def watch_copr_build(
        self,
        build_id: int,
        timeout: int,
        report_func: Optional[Callable] = None,
    ) -> str:
        """returns copr build state"""
        watch_end = datetime.now() + timedelta(seconds=timeout)
        logger.debug(f"Watching copr build {build_id}.")
        state_reported = ""
        while True:
            build = self.copr_client.build_proxy.get(build_id)
            if build.state == state_reported:
                continue
            state_reported = build.state
            logger.debug(f"COPR build {build_id}, state = {state_reported}")
            try:
                gh_state, description = COPR2GITHUB_STATE[state_reported]
            except KeyError as exc:
                logger.error(f"COPR gave us an invalid state: {exc}")
                gh_state, description = "error", "Something went wrong."
            if report_func:
                report_func(
                    gh_state,
                    description,
                    build_id=build.id,
                    url=self.copr_web_build_url(build),
                )
            if gh_state != "pending":
                logger.debug(f"State is now {gh_state}, ending the watch.")
                return state_reported
            if datetime.now() > watch_end:
                logger.error(f"The build did not finish in time ({timeout}s).")
                report_func("error", "Build watch timeout")
                return state_reported
            time.sleep(10)

    def get_copr_builds(self, number_of_builds: int = 5) -> list:
        """
        Get the copr builds of this project done by packit.
        :return: list of builds
        """
        client = CoprClient.create_from_config_file()

        projects = [
            project.name
            for project in reversed(client.project_proxy.get_list(ownername="packit"))
            if project.name.startswith(
                f"{self.upstream_local_project.namespace}-{self.upstream_local_project.repo_name}-",
            )
        ][:5]

        builds: list = []
        for project in projects:
            builds += client.build_proxy.get_list(
                ownername="packit",
                projectname=project,
            )

        logger.debug("Copr builds fetched.")
        return [(build.id, build.projectname, build.state) for build in builds][
            :number_of_builds
        ]

    @staticmethod
    @ttl_cache(maxsize=1, ttl=timedelta(hours=12).seconds)
    def get_available_chroots() -> list:
        """
        Gets available copr chroots. Uses cache to avoid repetitive url fetching.

        Returns:
            List of valid chroots.
        """

        client = CoprClient.create_from_config_file()
        return list(
            filter(
                lambda chroot: not chroot.startswith("_"),
                client.mock_chroot_proxy.get_list().keys(),
            ),
        )

    def get_build(self, build_id: int) -> dict:
        """
        Get build details from Copr.

        Arguments:
            build_id -- Copr build id.

        :return: Dict
        """
        return self.copr_client.build_proxy.get(build_id)

    def get_repo_download_url(self, owner: str, project: str, chroot: str) -> str:
        """Provide a link to yum repo for the particular chroot"""
        copr_proj = self.copr_client.project_proxy.get(
            ownername=owner,
            projectname=project,
        )
        try:
            return copr_proj["chroot_repos"][chroot]
        except KeyError as e:
            raise PackitCoprProjectException(
                f"There is no such target {chroot} in {owner}/{project}.",
            ) from e

    def get_chroots(
        self,
        owner: Optional[str] = None,
        project: Optional[str] = None,
        copr_project=None,
    ) -> set[str]:
        """
        Get chroots set on a specific project. Use either `owner`+`project` or
        directly `copr_project`.

        Args:
            owner: Owner of the Copr project.
            project: Name of the Copr project.
            copr_project: Already fetched Copr project via project proxy of a
                Copr client.

        Returns:
            Set of all enabled chroots on the requested Copr project.
        """
        if not copr_project:
            copr_project = self.copr_client.project_proxy.get(
                ownername=owner,
                projectname=project,
            )
        return set(copr_project.chroot_repos.keys())



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest

from packit.utils.decorators import fallback_return_value


class TestFallbackReturnValue:
    @pytest.mark.parametrize(
        "raise_exception, decorator_exceptions",
        [
            pytest.param(ValueError, ValueError, id="raised"),
            pytest.param(ValueError, KeyError, id="raised"),
        ],
    )
    def test_fallback_return_value(self, raise_exception, decorator_exceptions):
        fallback_value = "test_fallback_value"

        @fallback_return_value(fallback_value, exceptions=decorator_exceptions)
        def simple_function(exc=None):
            """Simple test function."""
            if exc is not None:
                raise exc
            return 42

        # `except` accepts both single exception or tuple of exceptions, to make testing easier
        # we will transform also single exception to tuple
        decorator_exceptions = (
            decorator_exceptions
            if isinstance(decorator_exceptions, tuple)
            else (decorator_exceptions,)
        )

        if raise_exception:
            if raise_exception in decorator_exceptions:
                assert simple_function(raise_exception) == fallback_value
            elif raise_exception not in decorator_exceptions:
                with pytest.raises(raise_exception):
                    simple_function(raise_exception)

        elif not raise_exception:
            assert simple_function() == 42



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os
import sys
from importlib.metadata import PackageNotFoundError

import pytest
from flexmock import flexmock

from packit.api import get_packit_version
from packit.utils import sanitize_branch_name, sanitize_version


def test_get_packit_version_not_installed():
    flexmock(sys.modules["packit.api"]).should_receive("version").and_raise(
        PackageNotFoundError,
    )
    assert get_packit_version() == "NOT_INSTALLED"


def test_get_packit_version():
    flexmock(sys.modules["packit.api"]).should_receive("version").and_return("0.1.0")
    assert get_packit_version() == "0.1.0"


@pytest.mark.parametrize(
    "to,from_,exp",
    (("/", "/", "."), ("/a", "/a/b", ".."), ("/a", "/c", "../a")),
)
def test_relative_to(to, from_, exp):
    assert os.path.relpath(to, from_) == exp


@pytest.mark.parametrize(
    "inp,exp,ver",
    (
        ("pr/123", "pr-123", "pr123"),
        ("🌈🌈🌈", "🌈🌈🌈", "🌈🌈🌈"),
        ("@#$#$%", "------", ""),
        ("pr+1", "pr-1", "pr+1"),
        ("pr^1", "pr-1", "pr^1"),
        ("pr~1", "pr-1", "pr~1"),
        ("pr-1", "pr-1", "pr.1"),
    ),
)
def test_sanitize_branch(inp, exp, ver):
    assert sanitize_branch_name(inp) == exp
    assert sanitize_version(inp) == ver



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from collections.abc import Iterable
from pathlib import Path
from typing import Optional, Union

from packit.exceptions import PackitCommandFailedError
from packit.utils import commands  # so we can mock utils
from packit.utils.logging import logger


class PkgTool:
    """
    Wrapper around fedpkg/centpkg/cbs.
    """

    def __init__(
        self,
        fas_username: Optional[str] = None,
        directory: Union[Path, str, None] = None,
        tool: str = "fedpkg",
        sig: Optional[str] = None,
    ):
        """
        Args:
            fas_username: FAS username (used for cloning)
            directory: operate in this dist-git repository
            tool: pkgtool to use (fedpkg, centpkg, centpkg-sig)
            sig: name of the SIG; used for adjusting the path during cloning
        """
        self.fas_username = fas_username
        self.directory = Path(directory) if directory else None
        self.tool = tool
        self.sig = sig

    def __repr__(self):
        return (
            "PkgTool("
            f"fas_username='{self.fas_username}', "
            f"directory='{self.directory}', "
            f"sig='{self.sig}', "
            f"tool='{self.tool}')"
        )

    def new_sources(
        self,
        sources: Optional[Iterable[Path]] = None,
        fail: bool = True,
        offline: bool = False,
    ):
        sources = sources or []
        if not self.directory.is_dir():
            raise Exception(f"Cannot access {self.tool} repository: {self.directory}")

        cmd = [self.tool, "new-sources"]
        if offline:
            cmd.append("--offline")
        sources_ = [str(source) for source in sources] if sources else []
        return commands.run_command_remote(
            cmd=cmd + sources_,
            cwd=self.directory,
            error_message="Adding new sources failed:",
            print_live=True,
            fail=fail,
        ).success

    def sources(self, fail: bool = True) -> bool:
        """Run the 'sources' command

        Args:
            fail: Raise an exception if the command fails

        Returns:
            True, if the command finished successfully, False otherwise.
        """
        return commands.run_command_remote(
            cmd=[self.tool, "sources"],
            cwd=self.directory,
            error_message="Downloading source files from the lookaside cache failed:",
            print_live=True,
            fail=fail,
        ).success

    def build(
        self,
        scratch: bool = False,
        nowait: bool = False,
        koji_target: Optional[str] = None,
        srpm_path: Optional[Path] = None,
    ) -> Optional[str]:
        """
        Run build in Koji.

        Args:
            scratch: scratch (temporary) build or not?
            nowait: False == wait for the build to finish
            koji_target: koji target to build in (`koji list-targets`)
            srpm_path: use selected SRPM for build, not dist-git repo & ref

        Returns:
            The 'stdout' of the build command.
        """
        cmd = [self.tool, "build"]
        if scratch:
            cmd.append("--scratch")
        if nowait:
            cmd.append("--nowait")
        if koji_target:
            cmd += ["--target", koji_target]
        if srpm_path:
            cmd += ["--srpm", str(srpm_path)]

        try:
            return commands.run_command_remote(
                cmd=cmd,
                cwd=self.directory,
                error_message="Submission of build to koji failed.",
                fail=True,
                output=True,
                print_live=True,
            ).stdout

        except PackitCommandFailedError as ex:
            # fail on the fedpkg side, the build is triggered
            if (
                "watch_tasks() got an unexpected keyword argument 'ki_handler'"
                in ex.stderr_output
            ):
                logger.info(
                    f"'{self.tool} build' crashed. It is a known issue: "
                    "the build is submitted in koji anyway.",
                )
                logger.debug(ex.stdout_output)
                return ""

            raise

    def clone(
        self,
        package_name: str,
        target_path: Union[Path, str],
        branch: Optional[str] = None,
        anonymous: bool = False,
    ):
        """
        clone a dist-git repo; this has to be done in current env
        b/c we don't have the keytab in sandbox
        """
        cmd = [self.tool]
        if self.fas_username:
            cmd += ["--user", self.fas_username]
        cmd += ["-q", "clone"]
        if branch:
            cmd += ["--branch", branch]
        if anonymous:
            cmd += ["--anonymous"]

        cmd += [
            f"{self.sig}/rpms/{package_name}" if self.sig else package_name,
            str(target_path),
        ]

        error_msg = (
            f"{self.tool} failed to clone repository {package_name}; "
            "please make sure that you are authorized to clone repositories "
            "from Fedora dist-git - this may require SSH keys set up or "
            "Kerberos ticket being active."
        )
        commands.run_command(cmd=cmd, error_message=error_msg)

    def verrel(self) -> str:
        """
        Run the `verrel` command.
        """
        return commands.run_command(
            cmd=[self.tool, "verrel"],
            cwd=self.directory,
            output=True,
            error_message=f"`{self.tool} verrel` was not successfull",
        ).stdout.strip()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""The 'dist-git' subcommand for Packit"""

import click

from packit.cli.dist_git_init import init


@click.group("dist-git")
def dist_git():
    """Subcommand to collect dist-git related functionality"""


dist_git.add_command(init)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import re
import shutil
from typing import Optional

import packit.upstream
from packit.actions import ActionName
from packit.config.common_package_config import MultiplePackages
from packit.distgit import DistGit

logger = logging.getLogger(__name__)


class ChangelogHelper:
    def __init__(
        self,
        upstream: "packit.upstream.Upstream",
        downstream: Optional[DistGit] = None,
        package_config: Optional[MultiplePackages] = None,
    ) -> None:
        self.up = upstream
        self.dg = downstream
        self.package_config = package_config

    @staticmethod
    def resolve_release_suffix(
        package_config: MultiplePackages,
        release_suffix: Optional[str] = None,
        default_release_suffix: bool = False,
    ) -> Optional[str]:
        """
        Resolves the release suffix value, since it can be set from multiple places
        and also overriden to the default one that is generated by packit.

        Args:
            package_config: Package config that is used as fallback.
            release_suffix: Release suffix that was passed from CLI.
            default_release_suffix: Override for using the default one that ensures
                correct NVR ordering.

        Returns:
            `None` if packit is to use the default method of generating release
            suffix, otherwise string containing the release suffix.
        """
        if default_release_suffix:
            # we want to use the default packit-generated release suffix
            release_suffix = None
        elif release_suffix is None:
            # we want to get release suffix from the configuration
            release_suffix = package_config.release_suffix
        return release_suffix

    def get_entry_from_action(
        self,
        version: Optional[str] = None,
        resolved_bugs: Optional[list[str]] = None,
        upstream_tag: Optional[str] = None,
        previous_version: Optional[str] = None,
    ) -> Optional[str]:
        """
        Runs changelog-entry action if present and returns string that can be
        used as a changelog entry.

        Args:
            version: version to be set in specfile
            resolved_bugs: List of bugs that are resolved by the update (e.g. [rhbz#123]).

        Returns:
            Changelog entry or `None` if action is not present.
        """
        resolved_bugs_str = " ".join(resolved_bugs) if resolved_bugs else ""
        env = self.package_config.get_package_names_as_env() | {
            "PACKIT_PROJECT_VERSION": version,
            "PACKIT_RESOLVED_BUGS": resolved_bugs_str,
            "PACKIT_PROJECT_UPSTREAM_TAG": upstream_tag or "",
            "PACKIT_PROJECT_PREVIOUS_VERSION": previous_version or "",
        }
        messages = self.up.actions_handler.get_output_from_action(
            ActionName.changelog_entry,
            env=env,
        )
        if not messages:
            return None

        return "\n".join(line.rstrip() for line in messages)

    @staticmethod
    def sanitize_entry(entry: str) -> str:
        # escape macro references and macro/shell/expression expansions
        # that could break spec file parsing
        entry = re.sub(r"(?<!%)%(?=(\w+|[{[(]))", "%%", entry)
        # prepend asterisk at the start of a line with a space in order
        # not to break identification of entry boundaries
        return re.sub(r"^[*]", " *", entry, flags=re.MULTILINE)

    def get_changelog_entry(
        self,
        default_changelog_entry: str,
        full_version: str,
        upstream_tag: str,
        resolved_bugs: Optional[list[str]],
    ):
        """
        Get the sanitized changelog entry to be added to the changelog.

        Args:
            default_changelog_entry: entry that will be used if there is nothing else configured
            full_version: version that is being proposed
            upstream_tag: new tag in upstream
            resolved_bugs: list of bugs that should be referenced
        """

        previous_specfile_version = None
        try:
            previous_specfile_version = self.dg.specfile.version
        except Exception:
            logger.debug(
                "We were not able to get the previous version from specfile in dist-git",
            )

        action_output = self.get_entry_from_action(
            version=full_version,
            resolved_bugs=resolved_bugs,
            upstream_tag=upstream_tag,
            previous_version=previous_specfile_version,
        )

        comment = action_output or default_changelog_entry
        if (
            self.package_config.copy_upstream_release_description
            # in pull_from_upstream workflow, upstream git_project can be None
            and self.up.local_project
            and self.up.local_project.git_project
        ):
            release_description = self.up.local_project.git_project.get_release(
                tag_name=upstream_tag,
                name=full_version,
            ).body
            if release_description:
                comment = release_description

        if not action_output and resolved_bugs:
            comment += "\n"
            for bug in resolved_bugs:
                comment += f"- Resolves: {bug}\n"

        return self.sanitize_entry(comment)

    def update_dist_git(
        self,
        full_version: str,
        upstream_tag: str,
        resolved_bugs: Optional[list[str]] = None,
    ) -> None:
        """
        Update the spec-file in dist-git:
        * Sync content from upstream spec-file.
        * Set 'Version'.
        * Add new entry in the %changelog section
          (if %autochangelog macro is not used).

        Copy the upstream spec-file as is if no spec-file is present in downstream.
        (E.g. for new packages)

        Args:
            full_version: Version to be set in the spec-file.
            upstream_tag: The commit messages after last tag and before this tag are used
                to update the changelog in the spec-file.
            resolved_bugs: List of bugs that are resolved by the update (e.g. [rhbz#123]).
        """
        default_changelog_entry = f"- Update to version {full_version}"
        try:
            comment = (
                self.get_changelog_entry(
                    default_changelog_entry,
                    full_version,
                    upstream_tag,
                    resolved_bugs,
                )
                if not self.dg.specfile.has_autochangelog
                else None
            )
            self.dg.set_specfile_content(
                self.up.specfile,
                full_version,
                comment=comment,
            )
        except FileNotFoundError as ex:
            # no downstream spec file: this is either a mistake or
            # there is no spec file in dist-git yet, hence warning
            logger.warning(
                f"Unable to find a spec file in downstream: {ex}, copying the one from upstream.",
            )
            # non-git upstream
            if not self.up.absolute_specfile_path:
                raise

            shutil.copy2(
                self.up.absolute_specfile_path,
                self.dg.get_absolute_specfile_path(),
            )
            comment = self.get_changelog_entry(
                default_changelog_entry,
                full_version,
                upstream_tag,
                resolved_bugs,
            )
            # set the specfile content now that the downstream spec file is present
            self.dg.set_specfile_content(
                self.up.specfile,
                full_version,
                comment=comment,
            )

    def _get_release_for_source_git(
        self,
        current_commit: str,
        update_release: bool,
        release_suffix: Optional[str],
    ) -> Optional[str]:
        old_release = self.up.specfile.expanded_release
        if release_suffix:
            return f"{old_release}.{release_suffix}"

        if not update_release:
            return None

        try:
            old_release_int = int(old_release)
            new_release = str(old_release_int + 1)
        except ValueError:
            new_release = str(old_release)

        return f"{new_release}.g{current_commit}"

    def prepare_upstream_using_source_git(
        self,
        update_release: bool,
        release_suffix: Optional[str],
    ) -> None:
        """
        Updates changelog when creating SRPM within source-git repository.
        """
        current_commit = self.up.local_project.commit_hexsha
        release_to_update = self._get_release_for_source_git(
            current_commit,
            update_release,
            release_suffix,
        )

        msg = self.get_entry_from_action()
        if not msg and update_release:
            msg = f"- Downstream changes ({current_commit})"
        self.up.specfile.release = release_to_update
        if msg is not None:
            self.up.specfile.add_changelog_entry(msg)

    def prepare_upstream_locally(
        self,
        version: str,
        commit: str,
        update_release: bool,
        release: str,
    ) -> None:
        """
        Updates changelog when creating SRPM within upstream repository.

        Args:
            version: Version to be set in the spec-file.
            commit: Commit to be set in the changelog.
            update_release: Whether to change Release in the spec-file.
            release: Release to be set in the spec-file.
        """
        self.up.specfile.version = version
        last_tag = self.up.get_last_tag()
        msg = self.get_entry_from_action(version=version)
        if not msg and last_tag and update_release:
            msg = self.up.get_commit_messages(after=last_tag)
        if not msg and update_release:
            # no describe, no tag - just a boilerplate message w/ commit hash
            # or, there were no changes b/w HEAD and last_tag, which implies last_tag == HEAD
            msg = f"- Development snapshot ({commit})"
        # instead of changing version, we change Release field
        # upstream projects should take care of versions
        if update_release:
            # Make sure the new release has a dist tag if it is not already included explicitly
            # (in release) or implicitly (in up.specfile.raw_release)
            if (
                release
                and not release.endswith("%{?dist}")
                and not self.up.specfile.raw_release.endswith("%{?dist}")
            ):
                release = f"{release}%{{?dist}}"
            logger.debug(f"Setting Release in spec to {release!r}.")
            self.up.specfile.release = release
        if msg is not None:
            self.up.specfile.add_changelog_entry(msg)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import copy
import json
from collections.abc import Mapping
from logging import getLogger
from typing import Any, Optional, Union

from marshmallow import (
    Schema,
    ValidationError,
    fields,
    post_dump,
    post_load,
    pre_load,
    validates_schema,
)
from marshmallow_enum import EnumField

from packit.actions import ActionName
from packit.config import (
    CommonPackageConfig,
    Config,
    Deployment,
    PackageConfig,
)
from packit.config.aliases import DEPRECATED_TARGET_MAP
from packit.config.commands import TestCommandConfig
from packit.config.common_package_config import MockBootstrapSetup
from packit.config.job_config import (
    JobConfig,
    JobConfigTriggerType,
    JobConfigView,
    JobType,
    get_default_jobs,
)
from packit.config.notifications import (
    FailureCommentNotificationsConfig,
    FailureIssueNotificationsConfig,
    NotificationsConfig,
    PullRequestNotificationsConfig,
)
from packit.config.requirements import LabelRequirementsConfig, RequirementsConfig
from packit.config.sources import SourcesItem
from packit.constants import (
    CHROOT_SPECIFIC_COPR_CONFIGURATION,
    FAST_FORWARD_MERGE_INTO_KEY,
)
from packit.sync import SyncFilesItem

logger = getLogger(__name__)


class StringOrListOfStringsField(fields.Field):
    """Field type expecting a string or a list"""

    def _serialize(self, value, attr, obj, **kwargs) -> list[str]:
        return [str(item) for item in value]

    def _deserialize(self, value, attr, data, **kwargs) -> list[str]:
        if isinstance(value, list) and all(isinstance(v, str) for v in value):
            return value
        if isinstance(value, str):
            return [value]
        raise ValidationError(f"Expected 'list[str]' or 'str', got {type(value)!r}.")


class SyncFilesItemSchema(Schema):
    """Schema for SyncFilesItem"""

    src = StringOrListOfStringsField()
    dest = fields.String()
    mkpath = fields.Boolean(default=False)
    delete = fields.Boolean(default=False)
    filters = fields.List(fields.String(), missing=None)


class FilesToSyncField(fields.Field):
    """
    Field type representing SyncFilesItem

    This is needed in order to handle entries which are strings, instead
    of a dict matching SyncFilesItemSchema.
    """

    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs) -> list[dict]:
        return SyncFilesItemSchema().dump(value)

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> SyncFilesItem:
        if isinstance(value, dict):
            return SyncFilesItem(**SyncFilesItemSchema().load(value))
        if isinstance(value, str):
            return SyncFilesItem(src=[value], dest=value)
        raise ValidationError(f"Expected 'dict' or 'str', got {type(value)!r}.")


class ActionField(fields.Field):
    """
    Field class representing Action.
    """

    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        return {action_name.value: val for action_name, val in value.items()}

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ) -> dict:
        if not isinstance(value, dict):
            raise ValidationError(f"'dict' required, got {type(value)!r}.")

        self.validate_all_actions(actions=list(value))
        return {ActionName(key): val for key, val in value.items()}

    def validate_all_actions(self, actions: list) -> None:
        """
        Validates all keys and raises exception with list of all invalid keys
        """
        invalid_actions = [
            action for action in actions if not ActionName.is_valid_action(action)
        ]

        if invalid_actions:
            raise ValidationError(f"Unknown action(s) provided: {invalid_actions}")


class NotProcessedField(fields.Field):
    """
    Field class to mark fields which will not be processed, only generates warning.
    Can be passed additional message via additional_message parameter.

    :param str additional_message: additional warning message to be displayed
    """

    def _serialize(self, value: Any, attr: str, obj: Any, **kwargs):
        raise NotImplementedError

    def _deserialize(
        self,
        value: Any,
        attr: Optional[str],
        data: Optional[Mapping[str, Any]],
        **kwargs,
    ):
        logger.warning(f"{self.name!r} is no longer being processed.")
        additional_message = self.metadata.get("additional_message")
        if additional_message:
            logger.warning(f"{additional_message}")


class SourceSchema(Schema):
    """
    Schema for sources.
    """

    path = fields.String(required=True)
    url = fields.String(required=True)

    @post_load
    def make_instance(self, data, **_):
        return SourcesItem(**data)


class PullRequestNotificationsSchema(Schema):
    """Configuration of commenting on pull requests."""

    successful_build = fields.Bool(default=False)

    @post_load
    def make_instance(self, data, **kwargs):
        return PullRequestNotificationsConfig(**data)


class FailureCommentNotificationsSchema(Schema):
    """Configuration of commenting on failures."""

    message = fields.String(missing=None)

    @post_load
    def make_instance(self, data, **kwargs):
        return FailureCommentNotificationsConfig(**data)


class FailureIssueNotificationsSchema(Schema):
    """Configuration of createing issues in upstream."""

    create = fields.Bool(default=True)

    @post_load
    def make_instance(self, data, **kwargs):
        return FailureIssueNotificationsConfig(**data)


class NotificationsSchema(Schema):
    """Configuration of notifications."""

    pull_request = fields.Nested(PullRequestNotificationsSchema)
    failure_comment = fields.Nested(FailureCommentNotificationsSchema)
    failure_issue = fields.Nested(FailureIssueNotificationsSchema)

    @post_load
    def make_instance(self, data, **kwargs):
        return NotificationsConfig(**data)


class TestCommandSchema(Schema):
    """Configuration of test command."""

    default_labels = fields.List(fields.String, missing=None)
    default_identifier = fields.String(missing=None)

    @post_load
    def make_instance(self, data, **kwargs):
        return TestCommandConfig(**data)


class LabelRequirementsSchema(Schema):
    present = fields.List(fields.String)
    absent = fields.List(fields.String)

    @post_load
    def make_instance(self, data, **kwargs):
        return LabelRequirementsConfig(**data)


class RequirementsSchema(Schema):
    """Configuration of test command."""

    label = fields.Nested(LabelRequirementsSchema, missing=None)

    @post_load
    def make_instance(self, data, **kwargs):
        return RequirementsConfig(**data)


class ListOrDict(fields.Field):
    """Field type expecting a List[str] or Dict[str, Dict[str, Any]]
    Union is not supported by marshmallow so we have to validate manually :(
    https://github.com/marshmallow-code/marshmallow/issues/1191
    """

    def is_dict(self, value) -> bool:
        if (
            not isinstance(value, dict)
            or not all(isinstance(k, str) for k in value)
            or not all(isinstance(v, dict) for v in value.values())
        ):
            return False
        return True


class DistGitBranches(ListOrDict):
    ERR_MESSAGE = (
        "Expected 'list[str]' or 'dict[str, dict[str, list]]', got {value!r} (type {type!r})."
        '\nExample -> "dist_git_branches": '
        '{{"fedora-rawhide": {{"fast_forward_merge_into": ["f40"]}}, epel9: {{}}}}}}'
    )

    def _deserialize(
        self,
        value,
        attr,
        data,
        **kwargs,
    ) -> Union[list[str], dict[str, dict[str, list[str]]], None]:
        if not (
            (isinstance(value, list) and all(isinstance(v, str) for v in value))
            or self.is_dict(value)
        ):
            raise ValidationError(
                self.ERR_MESSAGE.format(value=value, type=type(value)),
            )
        return value

    def is_dict(self, value) -> bool:
        if not super().is_dict(value):
            return False
        if isinstance(value, dict):
            if any(
                fast_forward_merge_into
                for fast_forward_merge_into in value.values()
                if not isinstance(fast_forward_merge_into, dict)
            ):
                raise ValidationError(
                    self.ERR_MESSAGE.format(value=value, type=type(value)),
                )
            if any(
                key
                for fast_forward_merge_into in value.values()
                for key in fast_forward_merge_into
                if key != FAST_FORWARD_MERGE_INTO_KEY
            ):
                raise ValidationError(
                    self.ERR_MESSAGE.format(value=value, type=type(value)),
                )
        return True


class Targets(ListOrDict):
    def is_dict(self, value) -> bool:
        if not super().is_dict(value):
            return False
        # check the 'attributes', e.g. {'distros': ['centos-7']} or
        # {"additional_modules": "ruby:2.7,nodejs:12", "additional_packages": []}
        for attr in value.values():
            for key, value in attr.items():
                # distros is a list of str
                if key == "distros":
                    if isinstance(value, list) and all(
                        isinstance(distro, str) for distro in value
                    ):
                        return True
                    raise ValidationError(
                        f"Expected list[str], got {value!r} (type {type(value)!r})",
                    )
                # chroot-specific configuration:
                if key in CHROOT_SPECIFIC_COPR_CONFIGURATION:
                    expected_type = CHROOT_SPECIFIC_COPR_CONFIGURATION[key].__class__
                    if isinstance(value, expected_type):
                        return True
                    raise ValidationError(
                        f"Expected {expected_type}, got {value!r} (type {type(value)!r})",
                    )
                raise ValidationError(f"Unknown key {key!r} in {attr!r}")
        return True

    def _deserialize(self, value, attr, data, **kwargs) -> dict[str, dict[str, Any]]:
        targets_dict: dict[str, dict[str, Any]]
        if isinstance(value, list) and all(isinstance(v, str) for v in value):
            targets_dict = {key: {} for key in value}
        elif self.is_dict(value):
            targets_dict = value
        else:
            raise ValidationError(
                f"Expected 'list[str]' or 'dict[str,dict]', got {value!r} (type {type(value)!r}).",
            )

        for target in targets_dict:
            if target in DEPRECATED_TARGET_MAP:
                logger.warning(
                    f"Target '{target}' is deprecated. Please update your configuration "
                    f"file and use '{DEPRECATED_TARGET_MAP[target]}' instead.",
                )
        return targets_dict


class JobMetadataSchema(Schema):
    """Jobs metadata.

    TODO: to be removed after deprecation period.
          Will end also dist-git-branch and
          dist_git_branch deprecation period.
    """

    _targets = Targets(missing=None, data_key="targets")
    timeout = fields.Integer()
    owner = fields.String(missing=None)
    project = fields.String(missing=None)
    dist_git_branches = DistGitBranches(missing=None)
    branch = fields.String(missing=None)
    scratch = fields.Boolean()
    list_on_homepage = fields.Boolean()
    preserve_project = fields.Boolean()
    use_internal_tf = fields.Boolean()
    additional_packages = fields.List(fields.String(), missing=None)
    additional_repos = fields.List(fields.String(), missing=None)
    bootstrap = EnumField(MockBootstrapSetup, missing=None)
    fmf_url = fields.String(missing=None)
    fmf_ref = fields.String(missing=None)
    fmf_path = fields.String(missing=None)
    skip_build = fields.Boolean()
    env = fields.Dict(keys=fields.String(), missing=None)
    enable_net = fields.Boolean(missing=False)
    tmt_plan = fields.String(missing=None)
    tf_post_install_script = fields.String(missing=None)
    tf_extra_params = fields.Dict(missing=None)
    module_hotfixes = fields.Boolean()

    @pre_load
    def ordered_preprocess(self, data, **_):
        for key in ("dist-git-branch", "dist_git_branch"):
            if key in data:
                logger.warning(
                    f"Job metadata key {key!r} has been renamed to 'dist_git_branches', "
                    f"please update your configuration file.",
                )
                data["dist_git_branches"] = data.pop(key)
        for key in ("targets", "dist_git_branches"):
            if isinstance(data.get(key), str):
                # allow key value being specified as string, convert to list
                data[key] = [data.pop(key)]

        return data


def validate_repo_name(value):
    """
    marshmallow validation for a repository name. Any
    filename is acceptable: No slash, no zero char.
    """
    if any(c in "/\0" for c in value):
        raise ValidationError("Repository name must be a valid filename.")
    return True


class CommonConfigSchema(Schema):
    """
    Common configuration options and methods for a package.
    """

    config_file_path = fields.String(missing=None)
    specfile_path = fields.String(missing=None)
    downstream_package_name = fields.String(missing=None)
    upstream_project_url = fields.String(missing=None)
    upstream_package_name = fields.String(missing=None, validate=validate_repo_name)
    paths = fields.List(fields.String())
    upstream_ref = fields.String(missing=None)
    upstream_tag_template = fields.String()
    archive_root_dir_template = fields.String()
    dist_git_url = NotProcessedField(
        additional_message="it is generated from dist_git_base_url and downstream_package_name",
        load_only=True,
    )
    dist_git_base_url = fields.String(mising=None)
    dist_git_namespace = fields.String(missing=None)
    allowed_gpg_keys = fields.List(fields.String(), missing=None)
    spec_source_id = fields.Method(
        deserialize="spec_source_id_fm",
        serialize="spec_source_id_serialize",
    )
    synced_files = fields.List(FilesToSyncField())
    files_to_sync = fields.List(FilesToSyncField())
    actions = ActionField(default={})
    create_pr = fields.Bool(default=True)
    sync_changelog = fields.Bool(default=False)
    create_sync_note = fields.Bool(default=True)
    patch_generation_ignore_paths = fields.List(fields.String(), missing=None)
    patch_generation_patch_id_digits = fields.Integer(
        missing=4,
        default=4,
        validate=lambda x: x >= 0,
    )
    notifications = fields.Nested(NotificationsSchema)
    copy_upstream_release_description = fields.Bool(default=False)
    sources = fields.List(fields.Nested(SourceSchema), missing=None)
    merge_pr_in_ci = fields.Bool(default=True)
    srpm_build_deps = fields.List(fields.String(), missing=None)
    identifier = fields.String(missing=None)
    packit_instances = fields.List(EnumField(Deployment), missing=[Deployment.prod])
    issue_repository = fields.String(missing=None)
    release_suffix = fields.String(missing=None)
    update_release = fields.Bool(default=True)
    upstream_tag_include = fields.String()
    upstream_tag_exclude = fields.String()
    prerelease_suffix_pattern = fields.String()
    prerelease_suffix_macro = fields.String(missing=None)
    upload_sources = fields.Bool(default=True)
    test_command = fields.Nested(TestCommandSchema)
    require = fields.Nested(RequirementsSchema)
    status_name_template = fields.String(missing=None)
    sync_test_job_statuses_with_builds = fields.Bool(default=True)

    # Former 'metadata' keys
    _targets = Targets(missing=None, data_key="targets")
    timeout = fields.Integer()
    owner = fields.String(missing=None)
    project = fields.String(missing=None)
    dist_git_branches = DistGitBranches(missing=None)
    branch = fields.String(missing=None)
    scratch = fields.Boolean()
    list_on_homepage = fields.Boolean()
    preserve_project = fields.Boolean()
    use_internal_tf = fields.Boolean()
    additional_packages = fields.List(fields.String(), missing=None)
    additional_repos = fields.List(fields.String(), missing=None)
    bootstrap = EnumField(MockBootstrapSetup, missing=None)
    fmf_url = fields.String(missing=None)
    fmf_ref = fields.String(missing=None)
    fmf_path = fields.String(missing=None)
    env = fields.Dict(keys=fields.String(), missing=None)
    enable_net = fields.Boolean(missing=False)
    allowed_pr_authors = fields.List(fields.String(), missing=None)
    allowed_committers = fields.List(fields.String(), missing=None)
    allowed_builders = fields.List(fields.String(), missing=None)
    tmt_plan = fields.String(missing=None)
    tf_post_install_script = fields.String(missing=None)
    tf_extra_params = fields.Dict(missing=None)
    module_hotfixes = fields.Boolean()

    # Image Builder integration
    image_distribution = fields.String(missing=None)
    # these two are freeform so that users can immediately use IB's new features
    image_request = fields.Dict(missing=None)
    image_customizations = fields.Dict(missing=None)
    copr_chroot = fields.String(missing=None)

    # Packaging tool used for interaction with lookaside cache
    pkg_tool = fields.String(missing=None)
    sig = fields.String(missing=None)
    version_update_mask = fields.String(missing=None)

    parse_time_macros = fields.Dict(missing=None)
    osh_diff_scan_after_copr_build = fields.Boolean(missing=True)

    csmock_args = fields.String(missing=None)

    @staticmethod
    def spec_source_id_serialize(value: CommonPackageConfig):
        return value.spec_source_id

    @staticmethod
    def spec_source_id_fm(value: Union[str, int]):
        """
        method used in spec_source_id field.Method
        If value is int, it is transformed int -> "Source" + str(int)

        ex.
        1 -> "Source1"

        :return str: prepends "Source" in case input value is int
        """
        if value:
            try:
                value = int(value)
            except ValueError:
                # not a number
                pass
            else:
                # is a number!
                value = f"Source{value}"
        return value

    @post_load
    def make_instance(self, data, **_):
        return CommonPackageConfig(**data)

    @post_dump(pass_original=True)
    def adjust_files_to_sync(
        self,
        data: dict,
        original: CommonPackageConfig,
        **kwargs,
    ) -> dict:
        """Fix the files_to_sync field in the serialized object

        B/c CommonPackageConfig.files_to_sync is a derived value, the meaning of the
        original configuration can be re-established only by checking the
        '_files_to_sync_used' attribute, and removing the 'files_to_sync' field from
        the serialized data if it's False.

        Args:
            data: Already serialized data.
            original: Original object being serialized.

        Returns:
            Modified serialized data with the 'files_to_sync' key removed if it was
            not used in the original object.
        """
        if not original._files_to_sync_used:
            data.pop("files_to_sync", None)
        return data


class JobConfigSchema(Schema):
    """
    Schema for processing JobConfig config data.
    """

    job = EnumField(JobType, required=True, attribute="type")
    trigger = EnumField(JobConfigTriggerType, required=True)
    skip_build = fields.Boolean()
    manual_trigger = fields.Boolean()
    labels = fields.List(fields.String(), missing=None)
    packages = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(CommonConfigSchema()),
    )
    package = fields.String(missing=None)

    # sidetag group identifier for downstream Koji builds and Bodhi updates
    sidetag_group = fields.String(missing=None)

    # packages that depend on this downstream Koji build to be tagged into
    # a particular sidetag group
    dependents = fields.List(fields.String(), missing=None)

    # packages whose downstream Koji builds are required to be tagged into
    # a particular sidetag group by this downstream Koji build or Bodhi update
    dependencies = fields.List(fields.String(), missing=None)

    @pre_load
    def ordered_preprocess(self, data, **_):
        for package, config in data.get("packages", {}).items():
            for key in ("targets", "dist_git_branches"):
                if isinstance(config, dict) and isinstance(config.get(key), str):
                    # allow key value being specified as string, convert to list
                    data["packages"][package][key] = [config.pop(key)]

        return data

    @validates_schema
    def specfile_path_defined(self, data: dict, **_):
        """Check if a 'specfile_path' is specified for each package

        The only time 'specfile_path' is not required, is when the job is a
        'test' job.

        Args:
            data: partially loaded configuration data.

        Raises:
            ValidationError, if 'specfile_path' is not specified when
            it should be.
        """
        # Note: At this point, 'data' is still a dict, but values are already
        # loaded, this is why 'data["type"]' is already a JobType and not a string,
        # and the package configs below are PackageConfig objects, not dictionaries.
        if (data["type"] == JobType.tests and data.get("skip_build")) or data.get(
            "specfile_path",
        ):
            return

        errors = {}
        package: str
        config: PackageConfig
        for package, config in data.get("packages", {}).items():
            if not config.specfile_path:
                errors[package] = [
                    "'specfile_path' is not specified or "
                    "no specfile was found in the repo",
                ]
        if errors:
            raise ValidationError(errors)

    @post_load
    def make_instance(self, data, **_):
        package = data.pop("package")
        job_config = JobConfig(**data)
        return JobConfigView(job_config, package) if package else job_config


class PackageConfigSchema(Schema):
    """
    Schema for processing PackageConfig config data.

    This class is intended to handle all the logic that is internal
    to the configuration and it is possible to be done while loading
    or dumping the configuration.

    This includes, for example, setting default values which depend on
    the value of other keys, or validating key values according to the
    value of other keys.

    It does not include setting the value of keys based on context
    *external* to the config file (if there is a spec-file in the current
    path, for example).
    """

    jobs = fields.Nested(JobConfigSchema, many=True)
    packages = fields.Dict(
        keys=fields.String(),
        values=fields.Nested(CommonConfigSchema()),
    )

    # list of deprecated keys and their replacement (new,old)
    deprecated_keys = (("upstream_package_name", "upstream_project_name"),)

    @pre_load
    def ordered_preprocess(self, data: dict, **_) -> dict:
        """Rename deprecated keys, and set defaults for 'packages' and 'jobs'

        Args:
            data: configuration dictionary as loaded from packit.yaml

        Returns:
            Transformed configuration dictionary with defaults
            for 'packages' and 'jobs' set.
        """
        # Create a deepcopy(), so that loading doesn't modify the
        # dictionary received.
        data = copy.deepcopy(data)
        data = self.rename_deprecated_keys(data)
        # Don't use 'setdefault' in this case, as we should expect
        # downstream_package_name only if there is no 'packages' key.
        if "packages" not in data:
            package_name = data.pop("downstream_package_name")
            paths = data.pop("paths", ["./"])
            data["packages"] = {
                package_name: {
                    "downstream_package_name": package_name,
                    "paths": paths,
                },
            }
        data.setdefault("jobs", get_default_jobs())
        # By this point, we expect both 'packages' and 'jobs' to be present
        # in the config.
        data = self.rearrange_packages(data)
        data = self.rearrange_jobs(data)
        data = self.process_job_triggers(data)
        logger.debug(f"Repo config after pre-loading:\n{json.dumps(data, indent=4)}")
        return data

    def rename_deprecated_keys(self, data: dict) -> dict:
        """
        Based on tuples stored in tuple cls.deprecated_keys, reassigns old keys values to new keys,
        in case new key is None and logs warning
        :param data: conf dictionary to process
        :return: processed dictionary
        """
        if not data:  # data is None when .packit.yaml is empty
            return data

        for new_key_name, old_key_name in self.deprecated_keys:
            old_key_value = data.get(old_key_name, None)
            if old_key_value:
                logger.warning(
                    f"{old_key_name!r} configuration key was renamed to {new_key_name!r},"
                    f" please update your configuration file.",
                )
                new_key_value = data.get(new_key_name, None)
                if not new_key_value:
                    # prio: new > old
                    data[new_key_name] = old_key_value
                del data[old_key_name]
        return data

    @staticmethod
    def rearrange_packages(data: dict) -> dict:
        """Update package objects with top-level configuration values

        Top-level keys and values are copied to each package object if
        the given key is not set in that object already.

        Remove these keys from the top-level and return a dictionary
        containing only a 'packages' and 'jobs' key.

        Args:
            data: configuration dictionary, before any of the leaves
                having been loaded.

        Returns:
            A re-arranged configuration dictionary.
        """
        # Pop 'packages' and 'jobs' in order for 'data'
        # to contain only keys other then these when it comes
        # to merging it bellow.
        packages = data.pop("packages")
        jobs = data.pop("jobs")
        for k, v in packages.items():
            # First set the defaults which are not inherited from
            # the top-level, in case they are not set yet.
            v.setdefault("downstream_package_name", k)
            # Inherit default values from the top-level.
            v.update(data | v)
        return {"packages": packages, "jobs": jobs}

    @staticmethod
    def rearrange_jobs(data: dict) -> dict:
        """Set the selected package config objects in each job, and set defaults
        according to the values specified on the level of job-objects (if any).

        Args:
            data: Configuration dict with 'packages' and 'jobs' already in place.

        Returns:
            Configuration dict where the package objects in jobs are correctly set.
        """
        packages = data["packages"]
        jobs = data["jobs"]
        errors = {}
        for i, job in enumerate(jobs):
            # Validate the 'metadata' field if there is any, and merge its
            # content with the job.
            # Do this here in order to avoid complications further in the
            # loading process.
            if metadata := job.pop("metadata", {}):
                logger.warning(
                    "The 'metadata' key in jobs is deprecated and can be removed. "
                    "Nest config options from 'metadata' directly under the job object.",
                )
                schema = JobMetadataSchema()
                if errors := schema.validate(metadata):
                    raise ValidationError(errors)
                if not_nested_metadata_keys := set(schema.fields).intersection(job):
                    raise ValidationError(
                        f"Keys: {not_nested_metadata_keys} are defined outside job metadata "
                        "dictionary. Mixing obsolete metadata dictionary and new job keys "
                        "is not possible. Remove obsolete nested job metadata dictionary.",
                    )
                job.update(metadata)

            top_keys = {}
            for key in JobConfigSchema().fields:
                if (value := job.pop(key, None)) is not None:
                    top_keys[key] = value

            selected_packages = top_keys.pop("packages", None)
            # Check that only packages which are defined on the top-level are selected.
            # Do this here b/c the code further down requires this to be correct.
            incorrect_packages = (
                set(selected_packages).difference(packages)
                if isinstance(selected_packages, (dict, list))
                else None
            )
            if incorrect_packages:
                errors[
                    f"jobs[{i}].packages"
                ] = f"Undefined package(s) referenced: {', '.join(incorrect_packages)}."
                continue

            # There is no 'packages' key in the job, so
            # the job should handle all the top-level packages.
            if not selected_packages:
                jobs[i] = top_keys | {
                    "packages": {k: v | job for k, v in packages.items()},
                }
            # Some top-level packages are selected to be
            # handled by the job.
            elif isinstance(selected_packages, list):
                jobs[i] = top_keys | {
                    "packages": {
                        k: v | job
                        for k, v in packages.items()
                        if k in selected_packages
                    },
                }
            # Some top-level packages are selected to be
            # handled by the job AND have some custom config.
            elif isinstance(selected_packages, dict):
                jobs[i] = top_keys | {
                    "packages": {
                        k: packages[k] | job | v for k, v in selected_packages.items()
                    },
                }
            else:
                errors[f"'jobs[{i}].packages'"] = [
                    f"Type is {type(selected_packages)} instead of 'list' or 'dict'.",
                ]

        if errors:
            # This will shadow all other possible errors in the configuration,
            # as the process doesn't even get to the validation phase.
            raise ValidationError(errors)

        return data

    @staticmethod
    def process_job_triggers(data: dict) -> dict:
        """
        Expands jobs with multiple triggers.

        For example a job with `trigger: commit | koji_build` will be expanded into two jobs
        with the same options except for `trigger`; the first one will have `trigger: commit`
        and the second one will have `trigger: koji_build`.
        """
        jobs = data["jobs"]
        i = len(jobs) - 1
        while i >= 0:
            triggers = [t.strip() for t in jobs[i].get("trigger", "").split("|")]
            if len(triggers) <= 1:
                i -= 1
                continue
            original_job = jobs.pop(i)
            for trigger in reversed(triggers):
                job = copy.deepcopy(original_job)
                job["trigger"] = trigger
                jobs.insert(i, job)
            i -= 1
        return data

    @post_load
    def make_instance(self, data: dict, **_) -> PackageConfig:
        return PackageConfig(**data)


class UserConfigSchema(Schema):
    """
    Schema for processing Config config data.
    """

    debug = fields.Bool()
    fas_user = fields.String()
    fas_password = fields.String()
    keytab_path = fields.String()
    redhat_api_refresh_token = fields.String()
    upstream_git_remote = fields.String()
    github_token = fields.String()
    pagure_user_token = fields.String()
    pagure_fork_token = fields.String()
    github_app_installation_id = fields.String()
    github_app_id = fields.String()
    github_app_cert_path = fields.String()
    authentication = fields.Dict()
    command_handler = fields.String()
    command_handler_work_dir = fields.String()
    command_handler_pvc_env_var = fields.String()
    command_handler_image_reference = fields.String()
    command_handler_k8s_namespace = fields.String()
    command_handler_pvc_volume_specs = fields.List(fields.Dict())
    command_handler_storage_class = fields.String()
    appcode = fields.String()
    kerberos_realm = fields.String()
    package_config_path = fields.String(default=None)
    koji_build_command = fields.String()
    pkg_tool = fields.String()
    repository_cache = fields.String(default=None)
    add_repositories_to_repository_cache = fields.Bool(default=True)
    default_parse_time_macros = fields.Dict(missing=None)

    @post_load
    def make_instance(self, data, **kwargs):
        return Config(**data)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Tests for Upstream class
"""
import os
import re
import shutil
import subprocess
import sys
from datetime import datetime, timedelta
from pathlib import Path

import pytest
from flexmock import flexmock
from ogr import GithubService

import packit
from packit.actions import ActionName
from packit.config import Config, get_local_package_config
from packit.exceptions import PackitSRPMException
from packit.local_project import LocalProjectBuilder
from packit.upstream import Archive, GitUpstream, SRPMBuilder
from packit.utils.commands import cwd
from packit.utils.repo import create_new_repo
from tests.spellbook import (
    EMPTY_CHANGELOG,
    NO_VERSION_TAG_IN_SPECFILE,
    UPSTREAM_MACRO_IN_SOURCE,
    build_srpm,
    get_test_config,
    initiate_git_repo,
)


def test_get_spec_version(upstream_instance):
    u, ups = upstream_instance
    assert ups.get_specfile_version() == "0.1.0"


@pytest.mark.parametrize(
    "tag, tag_template, expected_output",
    [
        pytest.param(
            "1.0.0",
            "{version}",
            "1.0.0",
            id="no_command-pure_version-valid_template",
        ),
        pytest.param(
            "test-1.0.0",
            "test-{version}",
            "1.0.0",
            id="no_command-valid_tag-valid_template",
        ),
    ],
)
def test_get_current_version(tag, tag_template, expected_output, upstream_instance):
    u, ups = upstream_instance
    flexmock(ups)
    ups.package_config.upstream_tag_template = tag_template
    # just to simulate current_vesrsion_command set/notset
    ups.should_receive("get_last_tag").and_return(tag)

    assert ups.get_current_version() == expected_output


@pytest.mark.parametrize(
    "m_v,exp",
    (
        ("1.1000.1000000", "1.1000.1000000"),
        (None, "0.1.0"),
        ("0.0.3", "0.1.0"),
        ("176", "176"),
    ),
)
def test_get_version(upstream_instance, m_v, exp):
    u, ups = upstream_instance
    flexmock(packit.upstream, get_upstream_version=lambda _: m_v)

    assert ups.get_version() == exp

    u.joinpath("README").write_text("\nEven better now!\n")
    subprocess.check_call(["git", "add", "."], cwd=u)
    subprocess.check_call(["git", "commit", "-m", "More awesome changes"], cwd=u)

    assert ups.get_current_version() == "0.1.0"


def test_get_version_macro(upstream_instance):
    u, ups = upstream_instance

    data = "import setuptools \nsetuptools.setup(version='1')"
    setup_path = u.joinpath("setup.py")
    with open(u.joinpath("setup.py"), "w+") as setup:
        setup.write(data)

    data = u.joinpath("beer.spec").read_text()
    data = data.replace("0.1.0", "%(python3 " + str(setup_path) + " --version)")
    with open(u.joinpath("beer.spec"), "w") as f:
        f.write(data)

    assert ups.get_specfile_version() == "1"


def test_get_version_no_version_tag(tmp_path):
    u_remote_path = tmp_path / "upstream_remote"
    u_remote_path.mkdir(parents=True, exist_ok=True)

    create_new_repo(u_remote_path, ["--bare"])

    u = tmp_path / "upstream_git"
    shutil.copytree(NO_VERSION_TAG_IN_SPECFILE, u)
    initiate_git_repo(u, tag="0.1.0")

    with cwd(tmp_path):
        c = get_test_config()

        pc = get_local_package_config(str(u))
        pc.upstream_project_url = str(u)
        lp = LocalProjectBuilder().build(working_dir=u)

        ups = GitUpstream(c, pc, lp)

    assert ups.get_specfile_version() == "2.1.1"


def test_set_spec_ver(upstream_instance):
    u, ups = upstream_instance

    new_ver = "1.2.3"
    ups.specfile.version = new_ver
    ups.specfile.add_changelog_entry("- asdqwe")

    assert ups.get_specfile_version() == new_ver
    assert "- asdqwe" in u.joinpath("beer.spec").read_text()


def test_set_spec_macro_source(tmp_path):
    u_remote_path = tmp_path / "upstream_remote"
    u_remote_path.mkdir(parents=True, exist_ok=True)

    create_new_repo(u_remote_path, ["--bare"])

    u = tmp_path / "upstream_git"
    shutil.copytree(UPSTREAM_MACRO_IN_SOURCE, u)
    initiate_git_repo(u, tag="0.1.0")

    with cwd(tmp_path):
        c = get_test_config()

        pc = get_local_package_config(str(u))
        pc.upstream_project_url = str(u)
        lp = LocalProjectBuilder().build(working_dir=u)

        ups = GitUpstream(c, pc, lp)

    expected_sources = ups.specfile.sources
    new_ver = "1.2.3"
    ups.specfile.version = new_ver
    ups.specfile.add_changelog_entry("- asdqwe")

    assert ups.get_specfile_version() == new_ver
    assert ups.specfile.sources == expected_sources

    expected_sources = ups.specfile.sources
    new_rel = "121"
    ups.specfile.release = new_rel

    assert ups.specfile.release == new_rel
    assert ups.specfile.sources == expected_sources


def test_set_spec_ver_empty_changelog(tmp_path):
    u_remote_path = tmp_path / "upstream_remote"
    u_remote_path.mkdir(parents=True, exist_ok=True)

    create_new_repo(u_remote_path, ["--bare"])

    u = tmp_path / "upstream_git"
    shutil.copytree(EMPTY_CHANGELOG, u)
    initiate_git_repo(u, tag="0.1.0")

    with cwd(tmp_path):
        c = get_test_config()

        pc = get_local_package_config(str(u))
        pc.upstream_project_url = str(u)
        lp = LocalProjectBuilder().build(working_dir=u)

        ups = GitUpstream(c, pc, lp)

    new_ver = "1.2.3"
    ups.specfile.version = new_ver
    ups.specfile.add_changelog_entry("- asdqwe")

    assert ups.get_specfile_version() == new_ver
    assert "%changelog" not in u.joinpath("beer.spec").read_text()


def change_source_ext(upstream, extension):
    with upstream.specfile.sources() as sources:
        for source in sources:
            source.location = (
                source.location.rstrip("".join(Path(source.location).suffixes))
                + extension
            )


@pytest.mark.parametrize(
    "extension",
    [".tar.gz", ".tar.bz2"],
    ids=[".tar.gz", ".tar.bz2"],
)
def test_create_archive(upstream_instance, extension):
    u, ups = upstream_instance
    change_source_ext(ups, extension)

    ups.create_archive()

    assert u.glob(f"*{extension}")

    u.joinpath("README").write_text("\nEven better now!\n")
    subprocess.check_call(["git", "add", "."], cwd=u)
    subprocess.check_call(["git", "commit", "-m", "More awesome changes"], cwd=u)

    ups.create_archive()

    # we enforce .tar.gz now
    assert len(list(u.glob("*.tar.gz"))) == 1


@pytest.mark.parametrize(
    "with_create_archive_action",
    (False, True),
)
def test_create_archive_spec_subdir(upstream_instance, with_create_archive_action):
    """test archive creation when an archive is in root, but spec is in a subdir"""
    u, ups = upstream_instance

    if with_create_archive_action:
        archive_path = u.joinpath("beer.tar.gz")
        ups.package_config.actions[ActionName.create_archive] = [
            [
                "git",
                "archive",
                "--output",
                str(archive_path),
                "--prefix",
                f"{u.name}/",
                "HEAD",
            ],
            ["echo", str(archive_path)],
        ]

    packaging_dir = u / "packaging"
    packaging_dir.mkdir()
    spec_path = packaging_dir / "some.spec"
    spec_path.write_text("asd")
    ups._specfile_path = spec_path

    ups.create_archive()

    archives_in_packaging_dir = list(packaging_dir.glob("*.tar.gz"))
    assert len(archives_in_packaging_dir) == 1

    if with_create_archive_action:
        assert archives_in_packaging_dir[0].is_symlink()
        assert len(list(ups.local_project.working_dir.glob("*.tar.gz"))) == 1


def test_create_uncommon_archive(upstream_instance):
    u, ups = upstream_instance
    change_source_ext(ups, ".cpio")
    ups.create_archive()
    # we enforce .tar.gz now
    assert len(list(u.glob("*.tar.gz"))) == 1


def test_fix_spec(upstream_instance):
    u, ups = upstream_instance

    ups.package_config.upstream_package_name = "beer"
    archive = ups.create_archive()
    ups.fix_spec(
        archive=archive,
        version="_1.2.3",
        commit="_abcdef123",
        release="1.20200710085501945230.master.0.g133ff39",
    )

    release = ups.specfile.expanded_release
    # 1.20200710085501945230.master.0.g133ff39
    assert re.match(r"\d\.\d{20}\.\w+\.\d+\.g\w{7}", release)


def test_fix_spec_persists(upstream_instance):
    """verify that changing specfile in fix_spec action persists"""
    _, upstream = upstream_instance
    upstream.package_config.actions = {
        ActionName.fix_spec: "sed -i 's/^Version:.*$/Version: 1.0.0/' beer.spec",
    }
    SRPMBuilder(upstream)._fix_specfile_to_use_local_archive(
        "archive.tar.gz",
        update_release=False,
        release_suffix="1.%dist",
    )

    assert upstream.specfile.version == "1.0.0"


def test_fix_spec_action_with_macros(upstream_instance):
    """verify that macros are properly initialized for a fix spec action"""
    _, upstream = upstream_instance
    upstream.package_config.actions = {
        ActionName.fix_spec: (
            "sed -i 's/^Release:.*$/Release: "
            "${PACKIT_RPMSPEC_RELEASE}%{?dist}/' beer.spec"
        ),
    }

    flexmock(sys.modules["packit.upstream"]).should_receive("datetime").and_return(
        flexmock(datetime=flexmock(now=flexmock(strftime=lambda f: "1234"))),
    )

    flexmock(upstream.local_project).should_receive("commit_hexsha").and_return("4321")
    flexmock(upstream).should_receive("get_spec_release").with_args(
        release_suffix="",
    ).and_return("1.1234.main.0.1221")

    flexmock(upstream.actions_handler).should_receive("with_action").with_args(
        action=ActionName.fix_spec,
        env={
            "PACKIT_PROJECT_VERSION": "0.1.0",
            "PACKIT_RPMSPEC_RELEASE": "1.1234.main.0.1221",
            "PACKIT_PROJECT_COMMIT": "4321",
            "PACKIT_PROJECT_ARCHIVE": "archive.tar.gz",
            "PACKIT_PROJECT_BRANCH": "main",
            "PACKIT_CONFIG_PACKAGE_NAME": "beer",
            "PACKIT_UPSTREAM_PACKAGE_NAME": "beerware",
            "PACKIT_DOWNSTREAM_PACKAGE_NAME": "beer",
        },
    )

    SRPMBuilder(upstream)._fix_specfile_to_use_local_archive(
        "archive.tar.gz",
        update_release=True,
        release_suffix="",
    )


@pytest.mark.parametrize(
    "spec_source_id, expected_line",
    [
        pytest.param(
            "Source",
            r"Source:\s*fixed-source-archive.tar.gz",
            id="Source",
        ),
        pytest.param(
            "Source0",
            r"Source:\s*fixed-source-archive.tar.gz",
            id="Source0",
        ),
        pytest.param(
            "Source100",
            r"Source100:\s*fixed-source-archive.tar.gz",
            id="Source100",
        ),
    ],
)
def test__fix_spec_source(upstream_instance, spec_source_id, expected_line):
    u, ups = upstream_instance

    data = u.joinpath("beer.spec").read_text()
    data = re.sub(r"Source0", "Source", data)
    data = re.sub(r"(Source:.*)", "\\1\nSource100: extra-sources.tar.gz", data)
    with open(u.joinpath("beer.spec"), "w") as f:
        f.write(data)

    ups.package_config.spec_source_id = spec_source_id
    ups._fix_spec_source("fixed-source-archive.tar.gz")

    assert re.search(expected_line, u.joinpath("beer.spec").read_text())


def test_create_srpm(upstream_instance, tmp_path):
    u, ups = upstream_instance

    with pytest.raises(PackitSRPMException) as exc:
        ups.create_srpm()
    # Creating an SRPM failes b/c the source archive is not present.
    assert "tar.gz: No such file or directory" in str(exc.value)

    ups.create_archive()
    srpm = ups.create_srpm()

    assert srpm.exists()

    srpm_path = tmp_path.joinpath("custom.src.rpm")

    ups.prepare_upstream_for_srpm_creation()
    srpm = ups.create_srpm(srpm_path=srpm_path)
    r = re.compile(r"^- Development snapshot \(\w{8}\)$")
    with ups.specfile.sections() as sections:
        for line in sections.changelog:
            if r.match(line):
                break
        else:
            raise AssertionError("Didn't find the correct line in the spec file.")
    assert srpm.exists()
    build_srpm(srpm)


def test_create_srpm_git_desc_release(upstream_instance):
    u, ups = upstream_instance
    u.joinpath("README").write_text("\nEven better now!\n")
    subprocess.check_call(["git", "add", "."], cwd=u)
    subprocess.check_call(["git", "commit", "-m", "More awesome changes"], cwd=u)

    ups.create_archive()
    ups.prepare_upstream_for_srpm_creation()
    srpm = ups.create_srpm()
    assert srpm.exists()
    build_srpm(srpm)
    assert re.match(
        r".+beer-0.1.0-1\.\d{20}\.\w+\.\d\.g\w{7}\.(fc\d{2}|el\d).src.rpm$",
        str(srpm),
    )

    with ups.specfile.sections() as sections:
        assert "- More awesome changes (Packit Test Suite)" in sections.changelog


def test_github_app(upstream_instance, tmp_path):
    u, ups = upstream_instance

    fake_cert_path = tmp_path / "fake-cert.pem"
    fake_cert_path.write_text("hello!")

    user_config_file_path = tmp_path / ".packit.yaml"
    user_config_file_path.write_text(
        "---\n"
        f"authentication:\n"
        f"    github.com:\n"
        f"        github_app_private_key_path: {fake_cert_path}\n"
        f"        github_app_id: qwe\n",
    )
    flexmock(os).should_receive("getenv").with_args("XDG_CONFIG_HOME").and_return(
        str(tmp_path),
    )
    ups.config = Config.get_user_config()
    assert (
        GithubService(
            github_app_private_key_path=str(fake_cert_path),
            github_app_id="qwe",
        )
        in ups.config.services
    )


def create_git_tag(u, tag_name, time_to_set, branch=None):
    timestamp = time_to_set.strftime("%Y-%m-%dT%H:%M:%S")
    # Set the environment variable to ensure the tag's timestamp
    env = {"GIT_COMMITTER_DATE": timestamp}

    if branch is not None:
        original_branch = (
            subprocess.check_output(["git", "branch", "--show-current"], cwd=u, env=env)
            .strip()
            .decode()
        )

    try:
        if branch is not None:
            try:
                subprocess.check_call(["git", "checkout", branch], cwd=u, env=env)
            except subprocess.CalledProcessError:
                subprocess.check_call(
                    ["git", "checkout", "--orphan", branch],
                    cwd=u,
                    env=env,
                )
        Path(u, "tags").write_text(tag_name)
        subprocess.check_call(["git", "add", "tags"], cwd=u, env=env)
        subprocess.check_call(
            ["git", "commit", "-m", f"Tag with {tag_name}"],
            cwd=u,
            env=env,
        )
        subprocess.check_call(
            ["git", "tag", "-a", "-m", f"Tag with {tag_name}", tag_name],
            cwd=u,
            env=env,
        )
    finally:
        if branch is not None:
            subprocess.check_call(["git", "checkout", original_branch], cwd=u, env=env)


@pytest.mark.parametrize(
    "tags, before, last_tag",
    [
        (["0.2.0"], None, "0.2.0"),
        (["0.2.0", "0.3.0"], None, "0.3.0"),
        (["0.2.0", "0.3.0"], "0.3.0", "0.2.0"),
        (["0.2.0", "0.3.0", "0.4.0"], "0.4.0", "0.3.0"),
    ],
)
def test_get_last_tag(upstream_instance, tags, before, last_tag):
    u, ups = upstream_instance
    now_time = datetime.now()

    # Tag more commits in the history
    for i, tag in enumerate(tags):
        create_git_tag(u, tag, now_time + timedelta(minutes=i + 1))

    assert ups.get_last_tag(before=before) == last_tag


@pytest.mark.parametrize(
    "tags, tags_branch1, tags_branch2, merged_ref, before, last_tag",
    [
        (["0.2.0"], [], [], None, None, "0.2.0"),
        (["0.2.0"], [], [], "branch1", None, None),
        (["0.2.0"], ["b1-1.0"], [], "branch1", None, "b1-1.0"),
        (["0.2.0"], ["b1-1.0"], ["b2-1.0.0"], "branch2", None, "b2-1.0.0"),
        (
            ["0.2.0", "0.3.0"],
            ["b1-1.0", "b1-2.1"],
            ["b2-1.0.0", "b2-1.2.3"],
            None,
            None,
            "b2-1.2.3",
        ),
        (
            ["0.2.0", "0.3.0"],
            ["b1-1.0", "b1-2.1"],
            ["b2-1.0.0", "b2-1.2.3"],
            "branch1",
            None,
            "b1-2.1",
        ),
        (
            ["0.2.0", "0.3.0"],
            ["b1-1.0", "b1-2.1"],
            ["b2-1.0.0", "b2-1.2.3"],
            "branch2",
            None,
            "b2-1.2.3",
        ),
        (
            ["0.2.0", "0.3.0"],
            ["b1-1.0", "b1-2.1"],
            ["b2-1.0.0", "b2-1.2.3"],
            None,
            "0.3.0",
            "0.2.0",
        ),
        (
            ["0.2.0", "0.3.0"],
            ["b1-1.0", "b1-2.1"],
            ["b2-1.0.0", "b2-1.2.3"],
            None,
            "b1-2.1",
            "b1-1.0",
        ),
        (
            ["0.2.0", "0.3.0"],
            ["b1-1.0", "b1-2.1"],
            ["b2-1.0.0", "b2-1.2.3"],
            None,
            "b2-1.2.3",
            "b2-1.0.0",
        ),
    ],
)
def test_get_last_tag_multiple_branches(
    upstream_instance,
    tags,
    tags_branch1,
    tags_branch2,
    merged_ref,
    before,
    last_tag,
):
    u, ups = upstream_instance
    now_time = datetime.now()

    # Tag more commits in the history
    for i, tag in enumerate(tags):
        create_git_tag(u, tag, now_time + timedelta(minutes=i + 1))
    for i, tag in enumerate(tags_branch1):
        create_git_tag(u, tag, now_time + timedelta(minutes=i - 60 + 1), "branch1")
    for i, tag in enumerate(tags_branch2):
        create_git_tag(u, tag, now_time + timedelta(minutes=i + 60 + 1), "branch2")

    ups._merged_ref = merged_ref
    assert ups.get_last_tag(before=before) == last_tag


@pytest.mark.parametrize(
    "tags, upstream_tag_include, upstream_tag_exclude, before, last_tag",
    [
        (["2.0.0", "3.0.0"], r"^2\..+", None, None, "2.0.0"),
        (["2.0.0", "2.1.0", "2.1.1", "2.2.0"], r"^.+\.1\..+", None, None, "2.1.1"),
        (
            [
                "2.0.0",
                "2.1.0",
                "2.2.0",
                "2.1.1",
            ],
            None,
            r"^.+\.1\..+",
            None,
            "2.2.0",
        ),
        (
            ["2.0.0", "2.1.0", "2.2.0", "2.1.1", "2.2.1"],
            r"^.+\.1\..+",
            None,
            "2.1.1",
            "2.1.0",
        ),
        (
            ["2.0.0", "2.1.0", "2.2.0", "2.1.1", "3.0.0"],
            r"^2\..+",
            r"^.+\.1\..+",
            None,
            "2.2.0",
        ),
    ],
)
def test_get_last_tag_matching_config(
    upstream_instance,
    tags,
    upstream_tag_include,
    upstream_tag_exclude,
    before,
    last_tag,
):
    u, ups = upstream_instance

    ups.package_config.upstream_tag_include = upstream_tag_include
    ups.package_config.upstream_tag_exclude = upstream_tag_exclude

    now_time = datetime.now()

    # Tag more commits in the history
    for i, tag in enumerate(tags):
        create_git_tag(u, tag, now_time + timedelta(minutes=i + 1))

    assert ups.get_last_tag(before=before) == last_tag


@pytest.mark.parametrize(
    "template, expected_output",
    [
        pytest.param("{upstream_pkg_name}-{version}", "beerware-0.1.0", id="default"),
        pytest.param(
            "{version}-{upstream_pkg_name}",
            "0.1.0-beerware",
            id="ver-pkg_name",
        ),
    ],
)
def test_get_archive_root_dir(template, expected_output, upstream_instance):
    u, ups = upstream_instance
    flexmock(packit.upstream.tarfile).should_receive("is_tarfile").and_return(False)
    ups.package_config.archive_root_dir_template = template

    archive = ups.create_archive()
    archive_root_dir = Archive(ups, ups.get_version()).get_archive_root_dir(archive)

    assert archive_root_dir == expected_output


def test_create_archive_not_create_symlink(upstream_instance):
    """test archive creation when an archive is in root, but spec is in a subdir
    and create_symlink=False"""
    u, ups = upstream_instance

    archive_path = u.joinpath("beer.tar.gz")
    ups.package_config.actions[ActionName.create_archive] = [
        [
            "git",
            "archive",
            "--output",
            str(archive_path),
            "--prefix",
            f"{u.name}/",
            "HEAD",
        ],
        ["echo", str(archive_path)],
    ]

    packaging_dir = u / "packaging"
    packaging_dir.mkdir()
    spec_path = packaging_dir / "some.spec"
    spec_path.write_text("asd")
    ups._specfile_path = spec_path

    ups.create_archive(create_symlink=False)

    archives = list(packaging_dir.glob("*.tar.gz"))
    assert len(archives) == 1
    assert not archives[0].is_symlink()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock

from packit.api import PackitAPI
from packit.cli import utils
from packit.cli.utils import get_packit_api
from packit.config import CommonPackageConfig, JobConfig
from packit.config.job_config import JobConfigTriggerType, JobType
from packit.constants import DISTGIT_INSTANCES
from packit.local_project import CALCULATE, LocalProjectBuilder
from packit.upstream import NonGitUpstream
from tests.spellbook import get_test_config, initiate_git_repo


def test_is_upstream(upstream_and_remote):
    upstream, _ = upstream_and_remote
    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            working_dir=upstream,
            git_repo=CALCULATE,
        ),
    )
    assert api.upstream_local_project
    assert not api.downstream_local_project
    assert api.upstream_local_project.working_dir == upstream


def test_is_downstream(distgit_and_remote):
    downstream, _ = distgit_and_remote
    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            working_dir=downstream,
            git_repo=CALCULATE,
        ),
    )
    assert api.downstream_local_project
    assert not api.upstream_local_project
    assert api.downstream_local_project.working_dir == downstream


def test_url_is_downstream():
    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            git_url="https://src.fedoraproject.org/rpms/packit",
            git_repo=CALCULATE,
        ),
    )
    assert api.downstream_local_project
    assert not api.upstream_local_project


def test_url_is_upstream():
    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            git_url="https://github.com/packit/ogr",
            git_repo=CALCULATE,
        ),
    )
    assert api.upstream_local_project
    assert not api.downstream_local_project


@pytest.mark.parametrize(
    "remotes,package_config,is_upstream",
    [
        (
            [],
            flexmock(
                upstream_project_url=None,
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [],
            flexmock(
                upstream_project_url="some-url",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [("origin", "https://github.com/packit/ogr.git")],
            flexmock(
                upstream_project_url="some-url",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [("origin", "https://github.com/packit/ogr.git")],
            flexmock(
                upstream_project_url="https://github.com/packit/ogr",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [("upstream", "https://github.com/packit/ogr.git")],
            flexmock(
                upstream_project_url="https://github.com/packit/ogr",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [("origin", "https://src.fedoraproject.org/rpms/ogr.git")],
            flexmock(
                upstream_project_url="https://github.com/packit/ogr",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
                downstream_package_name=None,
            ),
            False,
        ),
        (
            [("origin", "https://src.fedoraproject.org/rpms/python-ogr.git")],
            flexmock(
                upstream_project_url="https://github.com/packit/ogr",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
                downstream_package_name=None,
            ),
            False,
        ),
        (
            [("origin", "https://src.fedoraproject.org/rpms/python-ogr.git")],
            flexmock(
                upstream_project_url=None,
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
                downstream_package_name=None,
            ),
            False,
        ),
        (
            [("origin", "https://src.fedoraproject.org/fork/user/rpms/python-ogr.git")],
            flexmock(
                upstream_project_url=None,
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
                downstream_package_name=None,
            ),
            False,
        ),
        (
            [("origin", "git@github.com:user/ogr.git")],
            flexmock(
                upstream_project_url="https://github.com/packit/ogr",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [
                ("remote", "https://some.remote/ur/l.git"),
                ("origin", "git@github.com:user/ogr.git"),
            ],
            flexmock(
                upstream_project_url="https://github.com/packit/ogr",
                dist_git_instance=DISTGIT_INSTANCES["fedpkg"],
                synced_files=None,
            ),
            True,
        ),
        (
            [
                ("remote", "https://some.remote/ur/l.git"),
                ("origin", "git@github.com:user/ogr.git"),
            ],
            JobConfig(
                type=JobType.copr_build,
                trigger=JobConfigTriggerType.pull_request,
                packages={
                    "package": CommonPackageConfig(
                        upstream_project_url="https://github.com/packit/ogr",
                        dist_git_base_url=DISTGIT_INSTANCES["fedpkg"].url,
                        dist_git_namespace=DISTGIT_INSTANCES["fedpkg"].namespace,
                    ),
                },
            ),
            True,
        ),
    ],
)
def test_get_api(tmp_path, remotes, package_config, is_upstream):
    repo = tmp_path / "project_repo"
    repo.mkdir(parents=True, exist_ok=True)
    initiate_git_repo(repo, remotes=remotes)

    flexmock(utils).should_receive("get_local_package_config").and_return(
        package_config,
    )

    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            working_dir=str(repo),
            git_repo=CALCULATE,
        ),
    )

    if is_upstream:
        assert api.upstream_local_project
    else:
        flexmock(PackitAPI).should_receive("init_kerberos_ticket").once()
        assert api.downstream_local_project
        assert api.dg


def test_get_api_non_git_upstream(tmp_path):
    repo = tmp_path / "project_repo"
    repo.mkdir(parents=True, exist_ok=True)
    initiate_git_repo(
        repo,
        remotes=[
            ("origin", "https://src.fedoraproject.org/fork/user/rpms/python-ogr.git"),
        ],
    )

    flexmock(utils).should_receive("get_local_package_config").and_return(
        JobConfig(
            type=JobType.pull_from_upstream,
            trigger=JobConfigTriggerType.release,
            packages={
                "package": CommonPackageConfig(
                    dist_git_base_url=DISTGIT_INSTANCES["fedpkg"].url,
                    dist_git_namespace=DISTGIT_INSTANCES["fedpkg"].namespace,
                ),
            },
        ),
    )

    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            working_dir=str(repo),
            git_repo=CALCULATE,
        ),
        check_for_non_git_upstream=True,
    )

    assert isinstance(api.up, NonGitUpstream)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from typing import Any, Union


def ensure_str(inp: Union[bytes, str]) -> str:
    """decode bytes on input or just return the string"""
    # yes, this func doesn't have anything to do with exceptions
    # but it needs to be placed in a leaf module
    # utils import from exceptions hence it can't be there
    return inp if isinstance(inp, str) else inp.decode()


class PackitException(Exception):
    pass


class PackitCommandFailedError(PackitException):
    """A command failed"""

    def __init__(
        self,
        *args,
        stdout_output: Union[str, bytes],
        stderr_output: Union[str, bytes],
    ):
        super().__init__(*args)
        self.stdout_output = ensure_str(stdout_output)
        self.stderr_output = ensure_str(stderr_output)


class PackitConfigException(PackitException):
    pass


class PackitMissingConfigException(PackitConfigException):
    pass


class PackitCoprException(PackitException):
    pass


class PackitCoprProjectException(PackitCoprException):
    pass


class PackitCoprSettingsException(PackitException):
    """
    Raised when we can't edit the Copr project settings.

    Contains the info about fields that we want to edit.
    """

    def __init__(
        self,
        *args: object,
        fields_to_change: dict[str, tuple[Any, Any]],
    ) -> None:
        self.fields_to_change = fields_to_change
        super().__init__(*args)


class PackitInvalidConfigException(PackitConfigException):
    """provided configuration file is not valid"""


class PackitSRPMException(PackitException):
    """Problem with the SRPM"""


class PackitMergeException(PackitException):
    """Failed to merge PR into base branch"""


class PackitDownloadFailedException(PackitException):
    """Failed to download file"""


class PackitLookasideCacheException(PackitException):
    """Problem with lookaside cache"""


class PackitSRPMNotFoundException(PackitSRPMException):
    """SRPM created but not found"""


class PackitFailedToCreateSRPMException(PackitSRPMException):
    """Failed to create SRPM"""


class PackitRPMException(PackitException):
    """Problem with the RPM"""


class PackitRPMNotFoundException(PackitRPMException):
    """RPM created but not found"""


class PackitFailedToCreateRPMException(PackitRPMException):
    """Failed to create RPM"""


class PackitGitException(PackitException):
    """Operation with a git repo failed"""


class PackitNotAGitRepoException(PackitGitException):
    """Target directory is not a git repository as we expected"""


class PackitBodhiException(PackitException):
    """There was a problem while interacting with Bodhi"""


class ImageBuilderError(PackitException):
    """An error happened during the Image Builder interaction"""

    def __init__(self, errors):
        self.errors = errors


class ReleaseSkippedPackitException(PackitException):
    """The release of the new version has been skipped due to a failing check"""



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""Command to initialize a source-git repository"""

import logging
from typing import Optional

import click
import git

from packit.api import PackitAPI
from packit.cli.types import GitRepoParameter
from packit.cli.utils import cover_packit_exception
from packit.config import get_context_settings
from packit.config.config import pass_config

logger = logging.getLogger(__name__)


@click.command("init", context_settings=get_context_settings())
@click.argument("upstream_ref")
@click.argument("source_git", type=GitRepoParameter(from_ref_param="upstream_ref"))
@click.argument("dist_git", type=GitRepoParameter())
@click.option(
    "--upstream-url",
    help="""Git URL of the upstream repository. It is saved
    in the source-git configuration if it is specified.""",
)
@click.option(
    "--upstream-remote",
    help="""Name of the remote pointing to the upstream repository.
    If --upstream-url is not specified, the fetch URL of this remote
    is saved in the source-git configuration as the Git URL of the
    upstream project. Defaults to 'origin'.""",
)
@click.option(
    "--pkg-tool",
    help="""Name or path of the packaging tool used to work
    with sources in the dist-git repo. A variant of 'rpkg'.
    Defaults to 'fedpkg' or the tool configured in the Packit
    configuration.""",
)
@click.option(
    "--pkg-name",
    help="""The name of the package in the distro.
    Defaults to the directory name of DIST_GIT.""",
)
@pass_config
@cover_packit_exception
def source_git_init(
    config,
    dist_git: git.Repo,
    source_git: git.Repo,
    upstream_ref: str,
    upstream_url: Optional[str],
    upstream_remote,
    pkg_tool: Optional[str],
    pkg_name: Optional[str],
):
    """Initialize SOURCE_GIT as a source-git repo by applying downstream
    patches from DIST_GIT as Git commits on top of UPSTREAM_REF.

    SOURCE_GIT needs to be an existing clone of the upstream repository.

    UPSTREAM_REF is a tag, branch or commit from SOURCE_GIT.

    SOURCE_GIT and DIST_GIT are paths to the source-git and dist-git
    repos. Branch names can be specified, separated by colons.

    If a branch name is specified for SOURCE_GIT, the branch is checked
    out and reset to UPSTREAM_REF.

    If a branch name is specified for DIST_GIT, the branch is checked
    out before setting up the source-git repo. This branch is expected
    to exist.

    Each Git commit created in SOURCE_GIT will have a 'From-dist-git-commit'
    trailer to mark the hash of the dist-git commit from which it is created.

    To learn more about source-git, please check

        https://packit.dev/docs/source-git/

    Examples:

    \b
        $ packit source-git init v2.3.1 src/acl:rawhide rpms/acl:rawhide
        $ packit source-git init --pkg-tool centpkg v2.3.1 src/acl rpms/acl
    """
    logger.warning(
        "Generating source-git repositories is experimental, "
        "please give us feedback if it does things differently than you expect.",
    )
    api = PackitAPI(config=config, package_config=None)
    api.init_source_git(
        dist_git=dist_git,
        source_git=source_git,
        upstream_ref=upstream_ref,
        upstream_url=upstream_url,
        upstream_remote=upstream_remote,
        pkg_tool=pkg_tool or config.pkg_tool,
        pkg_name=pkg_name,
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import tempfile
from os import getenv
from pathlib import Path
from typing import Optional, Union

from packit.config import Config, RunCommandType
from packit.local_project import LocalProject
from packit.utils import commands

logger = logging.getLogger(__name__)

RUN_COMMAND_HANDLER_MAPPING: dict[RunCommandType, type["CommandHandler"]] = {}


def add_run_command(kls: type["CommandHandler"]):
    RUN_COMMAND_HANDLER_MAPPING[kls.name] = kls
    return kls


class CommandHandler:
    """Generic interface to handle different run_commands"""

    name: RunCommandType

    def __init__(
        self,
        config: Config,
        local_project: Optional[LocalProject] = None,
        working_dir: Optional[Path] = None,
    ):
        self.local_project = local_project
        self.config = config
        self._working_dir = working_dir

    @property
    def working_dir(self) -> Union[str, Path]:
        if self.local_project:
            return self.local_project.working_dir

        if self._working_dir:
            return self._working_dir

        return tempfile.mkdtemp()

    def run_command(
        self,
        command: list[str],
        return_output: bool = True,
        env: Optional[dict] = None,
        cwd: Union[str, Path, None] = None,
        print_live: bool = False,
    ) -> commands.CommandResult:
        """
        Run provided command in a new subprocess.

        Args:
            command: Command to be run.
            return_output: Return the output of the subprocess.

                Defaults to `False`.
            env: Environment variables to be set in the newly created subprocess.

                Defaults to none.
            cwd: Working directory of the new subprocess.

                Defaults to current working directory of the process itself.
            print_live: Print real-time output of the command as INFO.

                Defaults to `False`.

        Returns:
            In case `output = False`, returns boolean that indicates success of
            the command that was run.

            Otherwise returns pair of `stdout` and `stderr` produced by the
            subprocess. It can be pair of bytes or string, depending on the value
            of `decode` parameter.
        """
        raise NotImplementedError("This should be implemented")

    def clean(self) -> None:
        """clean up the mess after we're done"""
        logger.info("Nothing to clean.")


@add_run_command
class LocalCommandHandler(CommandHandler):
    """
    Local command handler that runs commands locally without any sandboxing.
    """

    name = RunCommandType.local

    def run_command(
        self,
        command: list[str],
        return_output: bool = True,
        env: Optional[dict] = None,
        cwd: Union[str, Path, None] = None,
        print_live: bool = False,
    ) -> commands.CommandResult:
        return commands.run_command(
            cmd=command,
            cwd=cwd or self.working_dir,
            output=return_output,
            env=env,
            print_live=print_live,
        )


@add_run_command
class SandcastleCommandHandler(CommandHandler):
    """
    Sandcastle command handler that runs commands in a sandbox provided by
    a Sandcastle.

    Warning: `.success` on returned `CommandResult` is set to `False`, since
    only logs are fetched from the Sandcastle pods.
    """

    name = RunCommandType.sandcastle

    def __init__(
        self,
        config: Config,
        local_project: Optional[LocalProject] = None,
        working_dir: Optional[Path] = None,
    ):
        super().__init__(
            local_project=local_project,
            config=config,
            working_dir=working_dir,
        )
        self.local_project = local_project
        self.config = config
        # we import here so that packit does not depend on sandcastle (and thus python-kube)
        from sandcastle.api import MappedDir, Sandcastle

        self._sandcastle: Optional[Sandcastle] = None
        self._mapped_dir: Optional[MappedDir] = None

    @property
    def sandcastle(self):
        """initialize Sandcastle lazily"""
        if self._sandcastle is None:
            # we import here so that packit does not depend on sandcastle (and thus python-kube)
            from sandcastle.api import MappedDir, Sandcastle, VolumeSpec

            self._mapped_dir = MappedDir(
                local_dir=self.config.command_handler_work_dir,  # path on worker
                path=self.config.command_handler_work_dir,  # path on sandcastle, it's the same!
                with_interim_pvc=True,
            )

            pvc_volume_specs = [
                VolumeSpec(**vol_spec_dict)
                for vol_spec_dict in self.config.command_handler_pvc_volume_specs
                if (
                    # Do not mount when the env-var is not set
                    "pvc_from_env" not in vol_spec_dict
                    or getenv(vol_spec_dict.get("pvc_from_env"))
                )
                and not (
                    # Do not mount repository cache when it's not used
                    self.config.repository_cache
                    and vol_spec_dict.get("path") == self.config.repository_cache
                    and not (
                        self.local_project
                        and self.local_project.cache.projects_cloned_using_cache
                    )
                )
            ]

            self._sandcastle = Sandcastle(
                image_reference=self.config.command_handler_image_reference,
                k8s_namespace_name=self.config.command_handler_k8s_namespace,
                mapped_dir=self._mapped_dir,
                volume_mounts=pvc_volume_specs,
                storage_class=self.config.command_handler_storage_class,
                appcode=self.config.appcode,
            )
            logger.debug("running the sandcastle pod")
            self._sandcastle.run()
        return self._sandcastle

    @staticmethod
    def _print_logs(logs: str) -> None:
        """
        Print out logs line-by-line using logger with the INFO level.

        Args:
            logs: Logs gathered from the Sandcastle.
        """
        # out = 'make po-pull\nmake[1]: Entering directory \'/sand
        for output_line in logs.split("\n"):
            if output_line:
                logger.info(output_line)

    def run_command(
        self,
        command: list[str],
        return_output: bool = True,
        env: Optional[dict] = None,
        cwd: Union[str, Path, None] = None,
        print_live: bool = False,
    ) -> commands.CommandResult:
        from sandcastle.exceptions import SandcastleCommandFailed

        cwd = cwd or Path(self.working_dir).relative_to(
            self.config.command_handler_work_dir,
        )
        logger.info(f"Running command: {' '.join(command)} on dir {cwd}")

        out = ""
        try:
            out = self.sandcastle.exec(command=command, env=env, cwd=cwd)

            logger.info(f"Output of {command!r}:")
            self._print_logs(out)
        except SandcastleCommandFailed as ex:
            # TODO: Maybe consider logging as an error? Would spam Sentry though
            logger.info(f"Command {command!r} failed:")
            self._print_logs(ex.output)
            raise  # reraise the error to halt the execution of the job

        return commands.CommandResult(
            success=True,
            stdout=out if return_output else None,
        )

    def clean(self) -> None:
        if self._sandcastle:
            logger.info("Deleting sandcastle pod.")
            self._sandcastle.delete_pod()
            self._sandcastle = None
        else:
            logger.info("Sandcastle pod is not running, nothing to clean up")

    def __del__(self) -> None:
        self.clean()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)
from packit.utils.changelog_helper import ChangelogHelper

logger = logging.getLogger("packit")


def build_rpms_from_specfile(api, upstream_ref, release_suffix, default_release_suffix):
    """
    Build RPMs from the specfile definition and return the paths to the built RPMs.
    """
    release_suffix = ChangelogHelper.resolve_release_suffix(
        api.package_config,
        release_suffix,
        default_release_suffix,
    )

    return api.create_rpms(
        upstream_ref=upstream_ref,
        release_suffix=release_suffix,
    )


def build_rpms_from_srpm(api, srpm):
    """
    Build RPMs from the SRPM and return the paths to the built RPMs.
    """
    return api.up.create_rpms_from_srpm(srpm)


def log_rpms(rpms):
    """
    Print out built RPMs after the RPM build is finished.
    """
    logger.info("RPMs:")
    for path in rpms:
        logger.info(f" * {path}")


@click.command("locally", context_settings=get_context_settings())
@click.option(
    "--upstream-ref",
    default=None,
    help="Git ref of the last upstream commit in the current branch "
    "from which packit should generate patches "
    "(this option implies the repository is source-git).",
)
@click.option(
    "--release-suffix",
    default=None,
    type=click.STRING,
    help="Specifies release suffix. Allows to override default generated:"
    "{current_time}.{sanitized_current_branch}{git_desc_suffix}",
)
@click.option(
    "--default-release-suffix",
    default=False,
    is_flag=True,
    help=(
        "Allows to use default, packit-generated, release suffix when some "
        "release_suffix is specified in the configuration."
    ),
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="build"),
)
@click.argument(
    "path_or_url",
    type=LocalProjectParameter(),
    default=os.path.curdir,
)
@pass_config
@cover_packit_exception
@iterate_packages
def local(
    config,
    upstream_ref,
    release_suffix,
    default_release_suffix,
    package_config,
    path_or_url,
):
    """
    Create RPMs using content of the upstream repository.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
    )

    rpms = (
        build_rpms_from_srpm(api, config.srpm_path)
        if config.srpm_path is not None
        else build_rpms_from_specfile(
            api,
            upstream_ref,
            release_suffix,
            default_release_suffix,
        )
    )

    log_rpms(rpms)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""    custom_path = "sooooorc.rpm"
Functional tests for srpm comand
"""
from pathlib import Path


def test_srpm(cwd_upstream_or_distgit, api_instance):
    u, d, api = api_instance
    api.create_srpm()
    assert next(Path.cwd().glob("*.src.rpm")).exists()


def test_srpm_custom_path(cwd_upstream_or_distgit, api_instance):
    u, d, api = api_instance
    custom_path = "sooooorc.rpm"
    api.create_srpm(output_file=custom_path)
    assert Path.cwd().joinpath(custom_path).is_file()



#!/usr/bin/python

# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import subprocess
from pathlib import Path

import fmf

tree_root = Path.cwd().absolute()
node = fmf.Tree(tree_root).find("/plans")
with node as data:
    data["discover"]["url"] = "https://github.com/packit/packit.git"
    data["discover"]["ref"] = (
        subprocess.check_output(["git", "rev-parse", "HEAD"]).decode().strip()
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from flexmock import flexmock
from gnupg import GPG

from packit.exceptions import PackitException
from packit.security import (
    VALID_SIGNATURE_STATUSES,
    CommitSignatureStatus,
    CommitVerifier,
)


@pytest.mark.parametrize(
    "sign,status",
    [
        ("N", CommitSignatureStatus.no_signature),
        ("B", CommitSignatureStatus.bad),
        ("G", CommitSignatureStatus.good_valid),
        ("U", CommitSignatureStatus.good_unknown_validity),
    ],
)
def test_commit_signature_status(sign, status):
    """Just to be sure we do not mess anything in the future."""
    assert status == CommitSignatureStatus(sign)


@pytest.mark.parametrize(
    "sign,status",
    [
        ("N", CommitSignatureStatus.no_signature),
        ("B", CommitSignatureStatus.bad),
        ("G", CommitSignatureStatus.good_valid),
        ("U", CommitSignatureStatus.good_unknown_validity),
    ],
)
def test_get_commit_signature_status(sign, status):
    """Just to be sure we do not mess anything in the future."""
    repo_mock = flexmock(git=flexmock().should_receive("show").and_return(sign).mock())

    status_found = CommitVerifier.get_commit_signature_status(
        commit=flexmock(hexsha="abcd", repo=repo_mock),
    )
    assert status_found == status


@pytest.mark.parametrize(
    "status,valid",
    [
        (CommitSignatureStatus.no_signature, False),
        (CommitSignatureStatus.bad, False),
        (CommitSignatureStatus.good_valid, True),
        (CommitSignatureStatus.good_unknown_validity, True),
    ],
)
def test_commit_signature_status_validity(status, valid):
    """Just to be sure we do not mess anything in the future."""
    is_valid = status in VALID_SIGNATURE_STATUSES
    assert is_valid == valid


@pytest.mark.parametrize(
    "key,sign,allowed_keys,local_keys,valid,download_times",
    [
        # good signature, key present
        ("a", "G", ["a"], ["a"], True, 0),
        # bad signature, key present
        ("a", "B", ["a"], ["a"], False, 0),
        # good signature, key not allowed
        ("a", "G", ["c", "d"], ["c", "d"], False, 0),
        # bad signature, key not present
        ("a", "B", ["c"], [], False, 0),
    ],
)
def test_check_signature_of_commit(
    key,
    sign,
    allowed_keys,
    local_keys,
    valid,
    download_times,
):
    gpg_flexmock = flexmock(GPG)
    gpg_flexmock.should_receive("list_keys").and_return(
        flexmock(fingerprints=local_keys),
    )
    gpg_flexmock.should_receive("recv_keys").and_return(
        flexmock(fingerprints=["fingerprint"]),
    ).times(download_times)

    repo_mock = flexmock(
        git=flexmock()
        .should_receive("show")
        .and_return(sign)
        .and_return(key)
        .and_return(sign)
        .mock(),
    )

    verifier = CommitVerifier()
    is_valid = verifier.check_signature_of_commit(
        commit=flexmock(hexsha="abcd", repo=repo_mock),
        possible_key_fingerprints=allowed_keys,
    )
    assert is_valid == valid


@pytest.mark.parametrize(
    "key,"
    "first_sign,"
    "second_sign,"
    "allowed_keys,"
    "local_keys,"
    "local_keys_after_download,"
    "valid,"
    "download_times",
    [
        # no signature
        (None, "N", None, [], [], None, False, 0),
        # key not present but downloaded
        ("a", "E", "G", ["a"], [], ["a"], True, 1),
        # key not present, all need to be downloaded before getting the signer
        ("a", "E", "G", ["a", "b", "c", "d"], [], ["a", "b", "c", "d"], True, 4),
        # key downloaded but signature is bad
        ("a", "E", "B", ["a"], [], ["a"], False, 1),
    ],
)
def test_check_signature_of_commit_not_present_key(
    key,
    first_sign,
    second_sign,
    allowed_keys,
    local_keys,
    local_keys_after_download,
    valid,
    download_times,
):
    gpg_flexmock = flexmock(GPG)
    gpg_flexmock.should_receive("list_keys").and_return(
        flexmock(fingerprints=local_keys),
    )

    gpg_flexmock.should_receive("recv_keys").and_return(
        flexmock(fingerprints=["fingerprint"]),
    ).times(download_times)

    repo_mock = flexmock(
        git=flexmock()
        .should_receive("show")
        .and_return(first_sign)
        .and_return(key)
        .and_return(second_sign)
        .mock(),
    )

    verifier = CommitVerifier()
    is_valid = verifier.check_signature_of_commit(
        commit=flexmock(hexsha="abcd", repo=repo_mock),
        possible_key_fingerprints=allowed_keys,
    )
    assert is_valid == valid


def test_check_signature_of_commit_key_not_found():
    gpg_flexmock = flexmock(GPG)

    # No key present
    gpg_flexmock.should_receive("list_keys").and_return(flexmock(fingerprints=[]))

    # No key received
    gpg_flexmock.should_receive("recv_keys").and_return(flexmock(fingerprints=[]))

    # Signature cannot be checked
    repo_mock = flexmock(git=flexmock().should_receive("show").and_return("E").mock())

    verifier = CommitVerifier()
    with pytest.raises(PackitException) as ex:
        verifier.check_signature_of_commit(
            commit=flexmock(hexsha="abcd", repo=repo_mock),
            possible_key_fingerprints=["a"],
        )
    assert "Cannot receive" in str(ex)


# This could possibly but unlikely fail if all the default key servers are down.
@pytest.mark.parametrize(
    "keyid, ok",
    [
        (  # Jirka's key id
            "A3E9A812AAB73DA7",
            True,
        ),
        # Don't, it times out if any key server is not responding
        # (
        #     "NOTEXISTING",
        #     False,
        # ),
    ],
)
def test_download_gpg_key_if_needed(keyid, ok):
    cf = CommitVerifier()
    if ok:
        assert cf.download_gpg_key_if_needed(keyid)
    else:
        with pytest.raises(PackitException):
            cf.download_gpg_key_if_needed(keyid)
